/* Copyright 1997 Lucent Technologies Inc. - All Rights Reserved */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#include <string.h>

#include "crypto.h"
#include "util.h"

/*
 * We only send and check half the digest.  It may make the
 * crypto a little harder to break.
 */
#define	SDIGEST	(sizeof(DIGEST)/2)
#define SDSTART	(sizeof(DIGEST) - SDIGEST - 1)

char sslerror[MAXIO];
char *sslname = "";

void
sslperror(char *msg) {
	fprintf(stderr, "%s: %s\n", msg, sslerror);
}

int
sslread(int fd, void *ibuf, int blen) {
	hmac_ctx ctx;
	DIGEST digest, sdigest;
	u_char count[2];
	u_char p[2];
	u_llong seq;
	u_char *mp, sbuf[sizeof(read_seq)];
	int len, n;

	n = netread(fd, (void *)count, 2);
	if (n < 0) {
		if (n == -2 || errno == ENOTCONN) {
			strcpy(sslerror, "EOF from network");
			return C_EOF;
		}
		sprintf(sslerror, "sslread failed: %s (%d)",
			strerror(errno), errno);
		return C_ERROR;
	}
	if(count[0] & 0x80)
		len = ((count[0] & 0x7f)<<8) | count[1];
	else {
		/*
		 * Padding is present, which we don't use.
		 */
		len = ((count[0] & 0x3f)<<8) | count[1];
		strcpy(sslerror,
			"Message format error, probably from an old version of the protocol");
		return C_ERROR;
	}
	if (len > blen) {
		sprintf(sslerror, "sslread: incoming record (%d) longer than expected (%d)",
			len, blen);
		return C_ERROR;
	}

	if (netread(fd, (void *)ibuf, len) < 0) {
		sprintf(sslerror, "sslread failed reading data: %s",
			strerror(errno));
		return C_ERROR;
	}
	if (netread(fd, (void *)sdigest, SDIGEST) < 0) {
		sprintf(sslerror, "sslread failed reading digest: %s",
			strerror(errno));
		return C_ERROR;
	}

	hmac_init(&ctx, (u_char *)&read_key, sizeof(read_key));
	hmac_update(&ctx, ibuf, len);
	read_seq++;
	mp = sbuf;
	*mp++ = read_seq>>54;
	*mp++ = read_seq>>48;
	*mp++ = read_seq>>40;
	*mp++ = read_seq>>32;
	*mp++ = read_seq>>24;
	*mp++ = read_seq>>16;
	*mp++ = read_seq>>8;
	*mp++ = read_seq;
	hmac_update(&ctx, sbuf, sizeof(sbuf));
	hmac_final(&ctx, digest);
	if (debug > 2) {
		fprintf(stderr, " read %5d", len);
		fprintf(stderr, "  seq %.16llx", read_seq);
		fprintf(stderr, "  key: %s", dk(read_key, 2));
		fprintf(stderr, "  dig %s", dd(digest, 4));
		fprintf(stderr, "  sdig %s", dd(sdigest, 4));
		fprintf(stderr, "  txt: %s",
			ds((u_char *)ibuf, len, 20));
	}

	if(memcmp(&digest[SDSTART], sdigest, SDIGEST) != 0) {
		if (strncmp(ibuf, ERRSTR, sizeof(ERRSTR)-1) == 0) {
			strncpy(sslerror, ibuf, sizeof(sslerror)-1);
			return -1;
		} else
			strcpy(sslerror, "corrupted message received");
		return C_CORRUPTED;
	}
	return len;
}

int
sslwrite(int fd, void *obuf, int len) {
	int m;
	DIGEST digest;
	hmac_ctx ctx;
	u_char buf[MAXIO+2+sizeof(SDIGEST)];
	u_char *mp, sbuf[sizeof(read_seq)];

	m = len;
	if (len > MAXIO) {
		fprintf(stderr, "sslwrite, block too large, %d > %d\n",
			len, 0x8000);
		return C_ERROR;
	}

	buf[0] = (m>>8) | 0x80;
	buf[1] = m & 0xff;

	hmac_init(&ctx, (u_char *)&write_key, sizeof(write_key));
	hmac_update(&ctx, obuf, len);
	write_seq++;
	mp = sbuf;
	*mp++ = write_seq>>54;
	*mp++ = write_seq>>48;
	*mp++ = write_seq>>40;
	*mp++ = write_seq>>32;
	*mp++ = write_seq>>24;
	*mp++ = write_seq>>16;
	*mp++ = write_seq>>8;
	*mp++ = write_seq;
	hmac_update(&ctx, (u_char *)&sbuf, sizeof(sbuf));
	hmac_final(&ctx, digest);

	memcpy(&buf[2], obuf, len);
	memcpy(&buf[2+len], &digest[SDSTART], SDIGEST);
	m = netwrite(fd, (void *)buf, 2+len+SDIGEST);
	if (m < 0) {
		strncpy(sslerror, strerror(errno), sizeof(sslerror)-1);
		return C_ERROR;
	}
	if (debug > 2) {
		fprintf(stderr, "wrote %5d", len);
		fprintf(stderr, "  seq %.16llx", write_seq);
		fprintf(stderr, "  key: %s", dk(write_key, 2));
		fprintf(stderr, "  dig %s", dd(digest, 4));
		fprintf(stderr, "  txt: %s",
			ds((u_char *)obuf, len, 20));
	}
	return len;
}
