/* Copyright 1997 Lucent Technologies Inc. - All Rights Reserved */

#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <sys/param.h>
#include <sys/file.h>
#include <fcntl.h>
#include <sys/time.h>
#include <ctype.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "get.h"
#include "client.h"
#include "crypto.h"
#include "util.h"
#include "arg.h"

int debug = 0;

int
usage(void) {
	fprintf(stderr,
		"usage: cget [-c cservfn] [-d] [-s] [-k keyfile] [-p port] [-h host] [-v] service remfn [locfn]\n");
	return(1);
}

int
main(int argc, char *argv[]) {
	char *hostname = 0;
	char *service = 0;
	char *getfile = 0;
	char *outfile = 0;
	char *keyfile = 0;
	char *host = 0;
	char *params;
	char *csrvfile = CSRVFILE;
	int serverport = SERVERPORT;
	int sflg = 0;
	struct stat st;
	KEY key;
	char *ep, *er;
	int srvfd, outfd, n;
	char buf[MAXIO];
	int bytes = 0, toreadbytes, mode;
	time_t	t0, t1;

	ARGBEGIN{
	case 'c': csrvfile = ARGF();		break;
	case 'd': debug++;			break;
	case 'h': hostname = ARGF();		break;
	case 'k': keyfile = ARGF();		break;
	case 'p': serverport = atoi(ARGF());	break;
	case 's': sflg++;			break;
	default:
		exit(usage());
	}ARGEND;

	if (argc < 2 || argc > 3)
		exit(usage());

	service = argv[0];
	getfile = argv[1];
	if (argc == 3)
		outfile = argv[2];
	else
		outfile = getfile;

	host = get_service_host(csrvfile, service);
	if (hostname == 0)
		hostname = host;
	get_key(keyfile, service, key);

	if ((srvfd = tcpconnect(hostname, serverport)) < 0) {
		fprintf(stderr, "cget: connection to %s:%d: %s\n",
			hostname, serverport, strerror(errno));
		exit(2);
	}

	ep = start_client_crypto(srvfd, key, service);
	if (ep) {
		fprintf(stderr, "cget: %s, %s: %s\n", hostname, service, ep);
		exit(3);
	}

	/* send the desired filename */
	sprintf(buf, "get %s", getfile);
	if (debug)
		fprintf(stderr, "sending '%s'\n", buf);
	cwrites(srvfd, buf);

	/*
	 * now get a response, which is either:
	 *	herecomes ###
	 * or an error
	 */
	if (cread(srvfd, buf, sizeof(buf)) < 0) {
		cperror("read server response");
		close(srvfd);
		exit(6);
	}

	if (debug)
		fprintf(stderr, "response: '%s'\n", buf);
	n = sscanf(buf, "herecomes %d %o %d",
		&toreadbytes, &mode, &st.st_mtime);
	if (n != 3) {
		fprintf(stderr,"%s\n",buf);
		close(srvfd);
		exit(7);
	}
	st.st_mode = mode;
	if (debug > 1)
		fprintf(stderr, "toread=%d mode=0%o mtime=%d\n",
			toreadbytes, mode, st.st_mtime);

	/*
	 * create the local file if not writing to stdout
	 */
	t0 = rtime();
	if (outfile == 0)
		outfd = 1;
	else {
		outfd = open(outfile, O_RDWR|O_CREAT|O_TRUNC, 0777 & st.st_mode);
		if(outfd < 0) {
			sprintf(buf,"cannot create %s as %s on local host",
				getfile, outfile);
			perror(buf);
			exit(8);
		}
		fchmod(outfd, 0777 & st.st_mode);
	}

	er = readnetfile(srvfd, outfd, toreadbytes);
	if (er) {
		fprintf(stderr, "can't get file: %s\n", er);
		unlink(getfile);
		exit(9);
	}
	close(outfd);

	/* final sync check */
	if (cread(srvfd, buf, sizeof(buf)) < 0) {
		cperror("read final sync");
		unlink(getfile);
		exit(11);
	}
	close(srvfd);
	if(strcmp(buf, "bye")) {
		fprintf(stderr, "file %s did not download properly\n",
			getfile);
		unlink(getfile);
		exit(12);
	}
	if(sflg) {
		double t;
		t1 = rtime();

		if (outfile == 0)
			outfile = "-";
		t = t1 - t0;
		fprintf(stderr, "cget: %s: %d bytes, %.2f secs, %.2f KB/s\n",
			outfile, toreadbytes, t/1000.0,
			toreadbytes/1024/(t/1000.0));
	}
	exit(0);
}
