/* Copyright 1997 Lucent Technologies Inc. - All Rights Reserved */

#include <ctype.h>
#include <stdio.h>
#include <fcntl.h>
#include <syslog.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <errno.h>
#include <sys/file.h>
#include <string.h>
#include <unistd.h>

#include "get.h"
#include "crypto.h"
#include "server.h"
#include "util.h"
#include "arg.h"

int debug = 0;
int unsafe = 0;

char *keydir = KEYDIR;
char *conffn = GETDCONF;

mode_t	srv_umask = 0;
mode_t	srv_or_perm = 0;

int	srv_readonly = 0;
/*int mode = DFLT_MODE;	/* file mode for cput */

int
usage(void) {
	fprintf(stderr, "usage: getserver [-v] [-d] [-c conffn] [-p port] [-k keydir]\n");
	return 1;
}


int
main(int argc, char * argv[]) {
	struct stat sbuf;
	char *caller;
	char filename[MAXPATHLEN], ifn[MAXPATHLEN];
	char buf[MAXIO];
	char *ep, *er, *cp;
	int ofd, fd;
	int x;
	int port = SERVERPORT;

	argv0 = argv[0];
	ARGBEGIN {
	case 'c': conffn = ARGF();	break;
	case 'd': debug++;		break;
	case 'k': keydir = ARGF();	break;
	case 'p': port = atoi(ARGF());	break;
	case 'U': unsafe++;		break;
	default:
		exit(usage());
	} ARGEND;

	if (debug) {
		logtostderr = 1;
		fd = debugbind(port);
		caller = getcaller(fd);
	} else {
		fd = 0;
		caller = getcaller(fd);
		if (caller == 0) {
			fprintf(stderr, "getd: input is not a socket\n");
			exit(1);
		}
		openlog("getd", LOG_PID, LOG_DAEMON);
	}

	/*
	 * If we don't make a syslog entry before we relinquish
	 * privs, subsequent syslog entries won't appear.
	 */
	log("called from %s", caller);

	/*
	 * start_server_crypto calls setup_service (in server.c) which fetches
	 * the global values we use.
	 *
	 * We also relinquish access, via chroot, setuid, and setgid,
	 * as early as possible, which is during this crypto setup.
	 */
	if ((ep=start_server_crypto(fd)) != 0) {
		log(ep);
		exit(2);
	}
	if (debug)
		log("crypto setup complete");


	if (debug)
		log("reading file infomation...");
	if (cread(fd, buf, sizeof(buf)) < 0) {
		log("error reading file name: %s", cerror);
		exit(4);
	}

	if (debug)
		log("file information: %s", buf);

	/* getting or putting? */
	if (!strncmp(buf,"get ",4)) {
		if ((cp = server_path(&buf[4], filename)) != 0) {
			log(cp);
			cwrites(fd, buf);
			exit(5);
		}
		if (stat(filename, &sbuf)) {
			sprintf(buf, "service %s, remote file not found: %s",
				sname, filename);
			log(buf);
			cwrites(fd, buf);
			exit(6);
		}
		if (S_ISDIR(sbuf.st_mode)) {
			sprintf(buf, "service %s, remote is a directory: %s",
				sname, filename);
			log(buf);
			cwrites(fd, buf);
			exit(6);
		}

		if ((ofd = open(filename, O_RDONLY)) < 0) {
			sprintf(buf, "service %s, server cannot open %s: %s",
				sname, filename, strerror(errno));
			log(buf);
			cwrites(fd, buf);
			exit(7);
		}

		/* send them the file size and stuff */
		sprintf(buf, "herecomes %d %o %d",
			(u_long)sbuf.st_size, (unsigned short)sbuf.st_mode,
			sbuf.st_mtime);
		cwrites(fd, buf);

		/*
		 * We just send them the whole file, not the
		 * number of bits promised.  This is wrong.
		 * But what do we do if the file length changes?
		 */
		er = writenetfile(ofd, fd, sbuf.st_size);
		if (er) {
			cwrites(fd, er);
			log("service '%s' sending '%s': %s\n",
				sname, filename, er);
		}
		cwrites(fd, "bye");
		log("service '%s' sent '%s' to %s length %d\n",
			sname, filename, caller, sbuf.st_size);
		close(0);
		exit(0);
	} else {  /* put */
		int toread, n;
		int bytes = 0;
		char *ss;
		mode_t mode;

		n = sscanf(buf, "put %o %d %d %100s",
			&mode, &sbuf.st_mtime, &sbuf.st_size,
			ifn);
		if (n != 4) {
			sprintf(buf, "put: invalid request (%d): %s", n, buf);
			cwrites(fd, buf);
			log(buf);
			exit(8);
		}
		if (srv_readonly) {
			sprintf(buf, "put: your access is read-only");
			cwrites(fd, buf);
			log(buf);
			exit(9);
		}
		if ((cp = server_path(ifn, filename)) != 0) {
			log(cp);
			cwrites(fd, buf);
			exit(10);
		}
		/*
		 * %u guarantees a unique local file name.  This is
		 * Andrew Hume's fault.
		 */
		if (ss = strstr(filename, "%u")) {
			char xx1[2*sizeof(filename)], xx2[2*sizeof(filename)];
			int xx;

			memcpy(xx1, filename, xx = ss-filename);
			xx1[xx] = 0;
			strcpy(xx2, ss+2);
			xx = getpid();
			for (;;){
				sprintf(buf, "%s%d%s", xx1, xx, xx2);
				if(access(buf, F_OK) < 0)
					break;
				xx++;
			}
			/* check for length here */
			strncpy(filename, buf, sizeof(filename)-1);
		}
		sbuf.st_mode = mode;
		if (((ofd = open(filename, O_WRONLY|O_CREAT|O_TRUNC, 0000))) < 0 &&
		     errno == EACCES) {
			struct stat sb;

			/*
			 * We couldn't open the file.  Perhaps it is an old
			 * mode 000 that we left around after a previous crash.
			 * If so, ``spit in the hole, man,  and tune again.''
			 */
			if ((stat(filename, &sb) == 0) &&
			   (sb.st_mode & (S_IRWXU|S_IRWXG|S_IRWXO)) == 0) {
				unlink(filename);
				ofd = open(filename,O_WRONLY|O_CREAT|O_TRUNC, 0000);
			}
		}
		if (ofd < 0) {
			sprintf(buf, "service %s, cannot create %s: %s",
				sname, filename, strerror(errno));
			log(buf);
			cwrites(fd, buf);
			exit(12);
		}
		/*
		 * while we are writing the file, the permissions are set to
		 * zero so others can check to see if the file is all here.
		 */
		chmod(filename, 0000);	/* clear permissions...we are working */
		cwrites(fd, "ok");

		toread = sbuf.st_size;
		/*
		 * We could check to see if there is room for this file.
		 */
		er = readnetfile(fd, ofd, toread);
		if (er) {
			char buf[1000];
			sprintf(buf, "service %s from %s, file read error: %s",
				sname, caller, er);
			log(buf);
			cwrites(fd, buf);
			unlink(filename);
			close(fd);  close(ofd);
			exit(15);
		}
		/*
		 * We are done writing the file.  `Unlock' it by setting its real mode.
		 */
		mode = ((sbuf.st_mode & 0777) & ~srv_umask) | srv_or_perm;
		fchmod(ofd, mode);
		close(ofd);
		cwrites(fd, "bye");
		log("service '%s' received '%s' mode %o length %d from %s",
			sname, filename, mode, toread, caller);
		close(0);
		exit(0);
	}
}
