/* Copyright 1997 Lucent Technologies Inc. - All Rights Reserved */

#include <unistd.h>
#include <ctype.h>
#include <stdio.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <errno.h>
#include <sys/file.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <utime.h>
#include <limits.h>
#include <syslog.h>

#include "stage.h"
#include "arg.h"
#include "server.h"
#include "crypto.h"
#include "util.h"

#define DFLT_MODE	0644

int debug = 0;
int vflg = 0;
int changes = 0;
int unsafe = 0;
int readonly = 0;

char *keydir = KEYDIR;
char *conffn = STAGEDCONF;

mode_t	srv_umask = UMASK;
mode_t	srv_or_perm = 0;

int	srv_readonly = 0;

int srv_fd;

enum {
	Remove,
	Stat,
	Checksum,
	Putfile,
	Getfile,
	Mkdir,
	Ls,
	Servinfo,
	Exit,
	Error
};

struct cmdtab {
	char *name;
	int c;
} cmd[] = {
	{"rm",	Remove},
	{"st",	Stat},
	{"cs",	Checksum},
	{"pu",	Putfile},
	{"ge",	Getfile},
	{"md",	Mkdir},
	{"ls",	Ls},
	{"si",	Servinfo},
	{"ex",	Exit},
	{0,	Error}
#define LASTCMD	((sizeof(cmd)/sizeof(struct cmdtab))-1)
};

void
reply(char *r) {
	int n = strlen(r)+1;
	if (vflg > 1)
		log("<--- %s", r);
	while (n > MAXIO) {
		if (cwrite(srv_fd, r, MAXIO) < 0) {
			log("error sending reply: %s", cerror);
			exit(1);
		}
		r += MAXIO, n -= MAXIO;
	}
	if (n > 0)
		if (cwrite(srv_fd, r, n) < 0) {
			log("error sending reply: %s", cerror);
			exit(1);
		}
}

void
senderr(char *fmt, void *p1) {
	char err[MAXPATHLEN];
	char buf[MAXIO];

	sprintf(err, fmt, p1);
	sprintf(buf, "NO %s", err);
	if (vflg > 1)
		log("<--e %s", buf);
	if (cwrite(srv_fd, buf, strlen(buf)+1) < 0) {
		log("error sending error: %s", cerror);
		exit(1);
	}
}

/*
 * Remove the thing at 'path'.  If it is a directory,
 * kill it and its contents. If an error has occurred,
 * send the protocol reply and return 0.  If all is well,
 * 1 is returned and the caller must reply.
 */
int
srv_remove(char *path) {
	struct stat st;
	char *subdir, *ep, *cp;
	int err=0;
	int rc = stat(path, &st);

	if (rc < 0) {
		if (errno == ENOENT)
			senderr("ENOENT", "");
		else {
			char buf[MAXIO];
			sprintf(buf, "remove: stat err %s %s",
				strerror(errno), path);
			senderr("%s", buf);
		}
		return 0;
	}

	if (S_ISREG(st.st_mode)) {
		if (unlink(path) == 0) {
			changes++;
			if (vflg)
				log("rm %s", unroot(path));
			return 1;
		} else {
			char buf[MAXIO];
			sprintf(buf, "%s %s", strerror(errno), path);
			senderr("%s", buf);
			return 0;
		}
	}

	/*
	 * run through the directory and kill its contents.
	 */
	subdir = dirls(path);
	for (cp=subdir; cp[1] != '/'; cp=ep) {
		char fn[MAXPATHLEN];

		ep = strchr(cp+1, '/');
		if (ep == 0)
			fprintf(stderr, "inconceivable ep\n");
		*ep = '\0';

		strcpy(fn, path);
		strcat(fn, cp);
		if (!srv_remove(fn)) {
			free(subdir);
			return 0;
		}
		*ep = '/';
	}
	free(subdir);
	if (rmdir(path) < 0) {
		char buf[MAXIO];
		sprintf(buf, "%d %s", errno, path);
		senderr("%s", buf);
		return 0;
	}
	if (vflg)
		log("rd %s", unroot(path));
	changes++;
	return 1;
}

void
srv_put(char *path) {
	struct stat *st;
	struct utimbuf tb;
	char buf[MAXIO];
	char *er;
	mode_t mode;
	int fd, i, j, k, n;

	if (cread(srv_fd, buf, sizeof(buf)) < 0) {
		senderr("protocol error reading stat in 'put'", cerror);
		return;
	}
	st = crackstat(&buf[3]);
	if (st == 0) {
		senderr("protocol error cracking stat in 'put'", "");
		return;
	}
	fd = open(path, O_CREAT|O_TRUNC|O_WRONLY, 0600);
	if (fd < 0) {
		senderr("%s", strerror(errno));
		return;
	}
	reply("OK");

	if (vflg > 1)
		log("---> (%d bytes)", st->st_size);
	er = readnetfile(srv_fd, fd, st->st_size);
	if (er) {
		senderr("could not copy file", er);
		close(fd);
		return;
	}
	mode = ((st->st_mode & 0777) & ~srv_umask) | srv_or_perm;
	if (chmod(path, mode) < 0) {
		senderr("could not set file mode", "");
		close(fd);
		return;
	}
	tb.modtime = st->st_mtime;
	tb.actime = st->st_mtime;
	if (utime(path, &tb))
		log("put: set times failed: %s", strerror(errno));
	close(fd);
	if (vflg)
		log("cr %s", unroot(path));
	changes++;
	reply("OK");
}

void
srv_get(char *path) {
	struct stat st;
	struct utimbuf tb;
	char buf[MAXIO];
	char *er;
	int fd, i, j, k, n;

	if (stat(path, &st) < 0) {
		senderr("%s", strerror(errno));
		return;
	}
	fd = open(path, O_RDONLY);
	if (fd < 0) {
		senderr("%s", strerror(errno));
		return;
	}
	reply("OK");
	reply(statinfo(path));

	if (vflg > 1)
		log("<--- (%d bytes)", st.st_size);
	er = writenetfile(fd, srv_fd, st.st_size);
	if (er) {
		senderr("server error reading file", er);
		close(fd);
		return;
	}
	close(fd);
	if (vflg)
		log("ge %s", unroot(path));
	changes++;
	reply("OK");
}

void
srv_mkdir(char *path) {
	struct stat *st;
	char buf[MAXIO];
	mode_t mode;

	if (cread(srv_fd, buf, sizeof(buf)) < 0) {
		senderr("protocol error reading stat in mkdir", cerror);
		return;
	}
	if (vflg > 1)
		log("---> %s", buf);
	st = crackstat(&buf[3]);
	if (st == 0) {
		senderr("protocol error cracking stat in mkdir", "");
		return;
	}

	mode = ((st->st_mode & 0777) & ~srv_umask) | srv_or_perm;
	if (path[0] != '\0' && mkdir(path, mode) < 0) {
		senderr("could not create directory", strerror(errno));
		return;
	}
	if (vflg)
		log("md %s", unroot(path));
	changes++;
	reply("OK");
}

int
usage(void) {
	fprintf(stderr, "usage: staged [-d] [-p port] [-u umask] [-v] [-c config] [-k keydir]\n");
	return 1;
}


int
main(int argc, char *argv[]) {
	char *ls, *caller, *cp, *ep;
	int port = SERVERPORT, n;
	int eof = 0;
	char path[MAXPATHLEN];
	char buf[MAXIO];

	ARGBEGIN {
	case 'c': conffn = ARGF();		break;
	case 'd': debug++;			break;
	case 'k': keydir = ARGF();		break;
	case 'p': port = atoi(ARGF());		break;
	case 'u': srv_umask = atoi(ARGF());	break;
	case 'U': unsafe++;			break;
	case 'v': vflg++;			break;
	default:
		exit(usage());
	} ARGEND;

	if (argc != 0)
		exit(usage());

	if (debug) {
		logtostderr = 1;
		srv_fd = debugbind(port);
		caller = getcaller(srv_fd);
		if (caller == 0) {
			fprintf(stderr, "staged: input is not a socket\n");
			exit(1);
		}
	} else {
		srv_fd = 0;
		caller = getcaller(srv_fd);
		if (caller == 0) {
			fprintf(stderr, "staged: input is not a socket\n");
			exit(1);
		}
		openlog("staged", LOG_PID, LOG_DAEMON);
	}

	/*
	 * If we don't make a syslog entry before we relinquish
	 * privs, subsequent syslog entries won't appear.
	 */
	log("called from %s", caller);

	/*
	 * start_server_crypto calls setup_service (in server.c)
	 * which fetches the global values we use.
	 *
	 * We also relinquish access, via chroot, setuid, and setgid,
	 * as early as possible, which is during the crypto setup
	 */
	if ((ep=start_server_crypto(srv_fd)) != 0) {
		log(ep);
		exit(2);
	}
	if (debug > 1)
		log("crypto setup complete");


	cmd[LASTCMD].name = buf;
	while (!eof && (n=cread(srv_fd, buf, sizeof(buf))) > 0) {
		struct cmdtab *cmdp = cmd;
		if (vflg > 1)
			log("---> %s", buf);

		while (cmdp->name && strncmp(buf, cmdp->name, 2) != 0)
			cmdp++;

		switch (cmdp->c) {
		case Remove:
			if (readonly) {
				reply("read only access");
				break;
			}
			if ((cp = server_path(&buf[3], path)) != 0) {
				reply(cp);
				break;
			}
			if (srv_remove(path))
				reply("OK");
			break;
		case Stat:
			if ((cp = server_path(&buf[3], path)) != 0) {
				reply(cp);
				break;
			}
			reply(statinfo(path));
			break;
		case Checksum:
			if ((cp = server_path(&buf[3], path)) != 0) {
				reply(cp);
				break;
			}
			reply(checksum(path));
			break;
		case Putfile:
			if (readonly) {
				reply("read only access");
				break;
			}
			if ((cp = server_path(&buf[3], path)) != 0) {
				reply(cp);
				break;
			}
			srv_put(path);
			break;
		case Getfile:
			srv_get(path);
			break;
		case Mkdir:
			if (readonly) {
				reply("read only access");
				break;
			}
			if ((cp = server_path(&buf[3], path)) != 0) {
				reply(cp);
				break;
			}
			srv_mkdir(path);
			break;
		case Ls:
			if ((cp = server_path(&buf[3], path)) != 0) {
				reply(cp);
				break;
			}
			ls = dirls(path);
			reply(ls);
			free(ls);
			break;
		case Servinfo:
			sprintf(buf, "%o %o",
				srv_umask, srv_or_perm);
			reply(buf);
			break;
		case Exit:
			reply("OK");
			close(srv_fd);
			eof = 1;
			break;
		case Error:
			senderr("proto: unknown command: %s\n", buf);
		}
	}
	if (!eof && n == C_ERROR)
		log("server read error: %s", cerror);
	log("Complete %s service '%s', %d change%c",
		caller, sname, changes, (changes == 1) ? '\0' : 's');
	close(srv_fd);
	exit(0);
}
