/* Copyright 1997 Lucent Technologies Inc. - All Rights Reserved */

#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/param.h>
#include <fcntl.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <limits.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <utime.h>

#include "arg.h"
#include "stage.h"
#include "client.h"
#include "crypto.h"
#include "util.h"

char	root[MAXPATHLEN];	/* Where the root of his file system is, locally */
char	*av0;

int debug = 0;	/* debug */
int nflg = 0;	/* don't make any changes */
int qflg = 0;	/* quick: skip cryptographic checksum */
int sflg = 0;	/* show stats */
int vflg = 0;	/* verbose - show protocol */
int only = 0;	/* only update directories, not their subdirectories */

int changes = 0;

int cmd_count = 0;	/* commands processed */
int files_checked = 0;
int files_transferred = 0;
int bytes_transferred = 0;
int responses = 0;
int response_wait = 0;

char *
get_root(void) {
	char *ep;
	char *cp = strtok(NULL, "	 \r\n");

	if (cp == NULL)
		return "/";

	/*
	 * Remove a trailing /, if any.
	 */
	ep = strrchr(root, '/');
	if (ep && (ep-root+1) == strlen(root))
		*ep = '\0';
	return cp;
}

/*
 * Delete a local directory or file, if it exists and !nflg.
 */
int
delete(char *path) {
	struct stat st;
	char *subdir, *ep, *cp;
	int err=0;
	int rc = stat(path, &st);

	if (rc < 0) {
		if (errno != ENOENT) {
			perror("delete");
			return 0;
		}
		fprintf(stderr, "could not file %s, file not present\n", path);
		return 1;
	}

	if (S_ISREG(st.st_mode)) {
		if (nflg) {
			printf("rm %s\n", path);
			return 1;
		}
		if (unlink(path) == 0) {
			changes++;
			if (vflg)
				printf("rm %s\n", path);
			return 1;
		}
		perror("delete file");
		return 0;
	}

	/*
	 * run through the directory and kill its contents.
	 */
	subdir = dirls(path);
	for (cp=subdir; cp[1] != '/'; cp=ep) {
		char fn[MAXPATHLEN];

		ep = strchr(cp+1, '/');
		if (ep == 0) {
			fprintf(stderr, "inconceivable ep\n");
			exit(1);
		}
		*ep = '\0';

		strcpy(fn, path);
		strcat(fn, cp);
		delete(fn);
		*ep = '/';
	}
	free(subdir);
	if (nflg) {
		printf("rd %s\n", path);
		return 1;
	}
	if (rmdir(path) < 0) {
		perror("rmdir");
		return 0;
	}
	if (vflg)
		printf("rd %s\n", path);
	changes++;
	return 1;
}

int
getfile(int rfd, char *lpath, char *rpath) {
	int i, n, fd, len;
	struct stat *stp;
	struct utimbuf tb;
	char buf[MAXIO];
	mode_t mode;
	char *er;

	remcmd(rfd, "ge", rpath);
	remresp(rfd);
	if (!remok) {
		fprintf(stderr, "could not fetch file %s: %s\n",
			rpath, remerr);
		return 0;
	}
	remresp(rfd);
	stp = crackstat(response);
	if (stp == 0) { 
		fprintf(stderr, "getfile protocol error: %s\n", response);
		return 0;
	}
	fd = open(lpath, O_CREAT|O_TRUNC|O_WRONLY, stp->st_mode);
	if (fd < 0) {
		perror("create file");
		return 0;
	}
	er = readnetfile(rfd, fd, stp->st_size);
	if (er) {
		fprintf(stderr, "Error reading %s: %s, aborting\n",
				lpath, er);
		exit(13);
	}
	if (vflg > 1)
		fprintf(stderr, "---> (%d bytes)\n", stp->st_size);
	mode = ((stp->st_mode & 0777) & ~srv_umask) | srv_or_perm;
	if (chmod(lpath, mode) < 0) {
		fprintf(stderr, "could not set file '%s' mode 0%o", 
			lpath, mode);
		close(fd);
		return;
	}
	tb.modtime = stp->st_mtime;
	tb.actime = stp->st_mtime;
	if (utime(lpath, &tb))
		fprintf(stderr, "set times failed on %s: %s",
			lpath, strerror(errno));
	close(fd);

	remresp(rfd);
	if (!remok) {
		fprintf(stderr, "could not get file %s: %s\n",
			rpath, remerr);
		return 0;
	}
	files_transferred++;
	bytes_transferred += len;
	return 1;
}

/*
 * Update the local file, if needed.  Return 1 if we actually had
 * to do it, 2 if update failed.
 *
 * On entry, laststat has stat of remote file that may be updated.
 */
int
check_file_for_update(int fd, char *path) {
	char *mess;
	struct stat st;
	int rc;

	files_checked++;
	if (debug)
		fprintf(stderr, "updating file %s\n", path);
	if (stat(path, &st) < 0) {
		if (errno != ENOENT) {
			perror("check_file_for_update stat:");
			return 2;
		}
		mess = "cr";
		goto getfile;
	}

	if (S_ISDIR(st.st_mode)) {	/* local file is a directory */
		if (!nflg && !delete(path))
			return 2;
		mess = "rdf";
		goto getfile;
	}

	/*
	 * Both local and remote are a file.  Compare, first the easy stuff,
	 * then with a checksum to see if we need to update the local file.
	 *
	 * XXX: this code is nearly identical in stage and unstage, and
	 * should be combined.  So little time, so many pedestrians...
	 */
	if (st.st_mtime == laststat.st_mtime &&
	    st.st_size == laststat.st_size) {
		char *lcs;
		mode_t remmode =
			((laststat.st_mode & 0777) & ~srv_umask) | srv_or_perm;
		mode_t locmode =
			  ((st.st_mode & 0777) & ~srv_umask) | srv_or_perm;
		if (locmode != remmode) {
			/*
			 * Permissions have changed.  Be lazy, and recopy
			 * the file.
			 */
			mess = "cm";
			goto getfile;
		}

		if (qflg)
			return 0;
		/* 
		 * Files might be the same, checksum.
		 */
		remcmd(fd, "cs", remotepath(path));
		lcs = checksum(path); /* two checksums at once! */
		remresp(fd);
		if (!remok) {
			/*
			 * remote file disappeared.  Create it.
			 */
			mess = "cr";
			goto getfile;
		}
		if (strlen(lcs) == sizeof(DIGEST)*2 &&
		    strlen(response) == sizeof(DIGEST)*2 &&
		    strcmp(lcs, response) == 0)
			return 0;
	}
	mess = "cp";

getfile:
	if (!nflg) {
		if (!getfile(fd, path, remotepath(path)))
			return 2;
	}	
	if (vflg || nflg)
		printf("%s %s\n", mess, path);
	changes++;
	return 1;
}

/*
 * Update the remote directory recursively.  Return 1 if any
 * changes were needed to the directory or any directory below it.
 * Return 2 if any errors occurred.
 */
int
check_dir_for_update(int fd, char *path) {
	struct stat st;
	char *lp, *lls, *rls, *ep, *cp;
	char *mess = "";
	int changed = 0, err = 0, newdir = 0;

	rls = get_remote_ls(fd, path);
	if (rls == 0) {
		fprintf(stderr, "check_dir_for_update, rls failure\n");
		return 2;
	}
	if (stat(path, &st) < 0) {
		if (errno != ENOENT) {
			perror("check_dir_for_update stat");
			free(rls);
			return 2;
		}
		newdir = 1;	/* remote file/dir not present */
		mess = "md";
	} else if (S_ISREG(st.st_mode)) {
		/*
		 * If the local item is a file, remove it, and create a fresh
		 * directory.
		 */
		if (!nflg && !delete(path)) {
			free(rls);
			return 2;
		}
		mess = "rfd";
		newdir = 1;
		changed = 1;
	} else {
		/*
		 * The local directory exists.  Make a string listing
		 * the local directory entries and compare it with the
		 * master remote list. Delete local entries not found
		 * in the remote master.
		 */
		lls = dirls(path);
		for (cp=lls; cp[1] != '/'; cp=ep) {
			char nc;

			ep = strchr(cp+1, '/');
			if (ep == 0) {
				fprintf(stderr, "check_dir_for_update: inconceivable\n");
				exit(13);
			}
			nc = ep[1];	/* save next char */
			ep[1] = '\0';
			if (strstr(rls, cp) == NULL) { /* not found remotely */
				char fn[MAXPATHLEN];
				ep[0] = '\0';
				strcpy(fn, path);
				strcat(fn, cp);
				if (!delete(fn))
					err = 1;
				changed = 1;
				ep[0] = '/';
			}
			ep[1] = nc;
		}
		free(lls);
	}
	if (newdir) {
		if (!nflg) {
			struct stat *sp = get_remote_stat(fd, path);
			mode_t mode;

			if (sp == 0) {
				fprintf(stderr, "remote dir %s vanished\n",
					remotepath(path));
				free(rls);
				return 2;
			}
			mode = ((sp->st_mode & 0777) & ~srv_umask) | srv_or_perm;
			if (mkdir(path, mode) < 0) {
				perror("check_dir_for_update mkdir");
				free(rls);
				return 2;
			}
		}
		changed = 1;
		changes++;
		if (vflg || nflg)
			printf("%s %s\n", mess, path);
	}

	/*
	 * Now, update each entry that is supposed to be there.
	 */
	for (lp=rls; lp[1] && lp[1] != '/'; *ep = '/', lp = ep) {
		char fn[MAXPATHLEN];
		struct stat *stp;

		ep = strchr(lp+1, '/');
		*ep = '\0';
		strcpy(fn, path);
		strcat(fn, lp);
		stp = get_remote_stat(fd, fn);
		if (stp == 0) {
			fprintf(stderr, "remote %s disappeared\n",
				remotepath(fn));
			err = 1;
			continue;
		}
		if (S_ISDIR(stp->st_mode)) {
			if (!only)
				changed |= check_dir_for_update(fd, fn);
		} else
			changed |= check_file_for_update(fd, fn);
	}
	free(rls);
	return (changed|err);
}

int
usage(void) {
	fprintf(stderr,
		"usage: unstage [-c config] [-h host] [-k keyfile] [-p port] [-q] [-s] [-v] [-n] service [path1 path2 ...]\n");
	return(1);
}

int
main(int argc, char *argv[]) {
	time_t	t0, t1;
	int srvfd;
	char *hostname = 0;
	char *service = 0;
	int serverport = SERVERPORT;
	char wd[MAXPATHLEN];
	int rc = 0;
	char *csrvfile = CSRVFILE;
	char *keyfile = 0;
	char *ep, *host;
	KEY key;

	ARGBEGIN {
	case 'c': csrvfile = ARGF();		break;
	case 'd': debug++;			break;
	case 'h': hostname = ARGF();		break;
	case 'k': keyfile = ARGF();		break;
	case 'n': nflg++;			break;
	case 'o': only++;			break;
	case 'p': serverport = atoi(ARGF());	break;
	case 'q': qflg++;			break;
	case 's': sflg++;			break;
	case 'v': vflg++;			break;
	default:
		exit(usage());
	} ARGEND;

	if (argc < 1)
		exit(usage());

	service = *argv++;
	argc--;

	host = get_service_host(csrvfile, service);
	if (hostname == 0)
		hostname = host;

	strcpy(root, get_root());
	if(debug)
		fprintf(stderr, "root=%s\n", root);

	get_key(keyfile, service, key);

	if ((srvfd = tcpconnect(hostname,serverport)) < 0) {
		fprintf(stderr, "unstage: connecting to %s:%d, service %s: %s\n",
			hostname, serverport, service, strerror(errno));
		exit(2);
	}

	ep = start_client_crypto(srvfd, key, service);
	if (ep) {
		fprintf(stderr, "unstage: %s, %s: %s\n", hostname, service, ep);
		exit(3);
	}
	if (debug)
		log("crypto setup complete");

	if (sflg)
		t0 = rtime();

	/*
	 * Drop any special setuid/setgid privs they may have given us to
	 * read the key file.
	 */
	if (setuid(getuid()) < 0) {
		perror("unstage: setuid");
		exit(13);
	}
	if (setgid(getgid()) < 0) {
		perror("unstage: setgid");
		exit(13);
	}
	if (getcwd(wd, MAXPATHLEN) == 0) {
		fprintf(stderr, "unstage: getwd: %s\n", wd);
		exit(13);
	}
	if (argc == 0) {
		argv[0] = root;
		argc = 1;
	}

	get_server_info(srvfd);

	for (; argc; argv++, argc--) {
		struct stat st;
		char path[MAXPATHLEN*2 + 1];

		if (strlen(argv[0]) > MAXPATHLEN-1) {
			fprintf(stderr, "path too long: %s\n", argv[0]);
			continue;
		}
		if (argv[0][0] == '/')
			strncpy(path, argv[0], sizeof(path)-1);
		else {
			/*
			 * Convert relative path to absolute.  Directory
			 * '.' is ugly in the logs: don't include it.
			 */
			strcpy(path, wd);
			if (strcmp(argv[0], ".") != 0) {
				strcat(path, "/");
				strcat(path, argv[0]);
			}
		}
		if (get_remote_stat(srvfd, path))
			if (S_ISDIR(laststat.st_mode))
				rc |= check_dir_for_update(srvfd, path);
			else
				rc |= check_file_for_update(srvfd, path);
		else
			delete(path);
	}
	remcmd(srvfd, "ex", "");
	remresp(srvfd);

	if (sflg) {
		double t;
		t1 = rtime();

		t = t1 - t0;
		fprintf(stderr, "%.2f cmds/s,  %.2f files/s,  %d updates,  %.1f ms ave resp.\n",
			cmd_count/(t/1000.0), files_checked/(t/1000.0),
			files_transferred, 
			(double)response_wait/(double)responses);
	}
	return rc;
}
