/*=============================*/
/*           NETS              */
/*                             */
/*     a product of the 	   */
/*	Software Technology Branch */
/*  NASA, Johnson Space Center */
/*                             */
/* authors:           		   */
/*       Paul Baffes           */
/*       Steve Bayer           */
/*       Bryan Dulock          */
/*       Linda Jensen          */
/*       Chris Ortiz           */
/*       Gary Riley            */
/*       Robert Shelton        */
/*       Todd Phillips         */
/*=============================*/


/*
----------------------------------------------------------------------
  This is the main module for the neural net.  All of the main code   
  plus documentation on the system features lives here.

  A note on nomenclature: Most of the routine calls which you will 
  see in the code throughout these files will be "prefixed" by some 
  letters and an underscore.  That is one of my conventions for indicating
  which file contains the code for the subroutine in question. Because
  I have several files, this makes tracing problems and debugging
  files easier to accomplish.  

  This file is one of 14 source files which make up the back
  propagation code. These files are the following:

      activate.c              semi-linear activation function
      buildnet.c
      compile.c
      convert.c               conversion routines, conversion routines
      layer.c                 layer manipulation and creation
      lnrate.c
      net.c                   net manipulation, learning, propagation
      netio.c                 I/O routines, file handlers
      netmain.c               main routines, menus, user interface
      pairs.c
      prop.c
      shownet.c
      scaleio.c				  Scales IO pairs and Input files
      teach.c
      weights.c               weights manipulation and creation

  All of these are covered in detail in the respective files. The prefix
  codes for each of the files are as follows:

      activate.c              "A_"
      buildnet.c              "B_"
      compile.c               "CC_"
      convert.c               "C_"
      dribble.c               "D_"
      net.c                   "N_"
      netio.c                 "IO_"
      netmain.c               *NONE*
      pairs.c                 "PA_"
      parser.c                "PS_"
      prop.c                  "P_"
      layer.c                 "L_"
      lnrate.c                "LR_"
      shownet.c               "S_"
      scaleio.c				  "SC_"
      teach.c                 "T_"
      weights.c               "W_"
      sysdep.c                "sys_"    (system dependent code)

  The rest of this file is organized into the folloing groups:

  (1) include files
  (2) externed functions
  (3) global variables
  (4) subroutines
----------------------------------------------------------------------
*/


/*
----------------------------------------------------------------------
  INCLUDE FILES
----------------------------------------------------------------------
*/
#include  "common.h"
#include  "netio.h"
#include  "weights.h"
#include  "layer.h"
#include  "net.h"

/* there are two versions of the main loop -- one for the text based interface and one for macintosh */

#if !MAC_GRAPH					
/*
----------------------------------------------------------------------
  EXTERNED FUNCTIONS
----------------------------------------------------------------------
*/
extern void		A_initialize();

extern Net    *B_create_net();
extern Net    *B_free_net();

extern void   N_query_net();
extern int    N_reset_wts();
extern void   N_save_wts();
extern Layer  *N_get_layer();

extern void T_signal_handle ();
extern void   T_teach_net();
extern void T_display_error_stats();

extern void   S_show_weights();
extern void   S_show_biases();
extern void   S_show_net();

extern float  IO_my_get_float();
extern int    IO_my_get_int();
extern int    IO_get_default_int();
extern int    IO_get_num_cycles();
extern void   IO_my_get_string();
extern void   IO_set_filenames();
extern void   IO_get_io_name();
extern void   IO_get_wts_name();
extern void   IO_print();

extern void   PA_initialize();
extern void   PA_setup_iopairs();
extern void   PA_reset_iopairs();
extern void   PA_randomize_file();
extern void PA_done_with_workfile();

extern void   D_initialize();
extern void   D_dribble_status(); 

extern Sint   C_float_to_Sint();
extern void   L_modify_learning();
extern void   sys_init_rand();
extern void	  sys_SetUpVBLTask();
extern void	  sys_RemoveVBLTask();
extern void   CC_create_delivery();


/*
----------------------------------------------------------------------
  GLOBAL VARIABLES
----------------------------------------------------------------------
 Next come the global variables, declared in other routines, which    
  need to be referenced here.  In general I tried to keep the number  
  of globals to a minimum since they can be messy, but many of the    
  io functions needed to keep "state" variables for the lifetime of   
  the program execution.  Examples are the default file names which   
  are used when prompting the user.  These names are read here, and   
  referenced in  net.c as well as netio.c, and thus I needed to be    
  able to pass them around.  I could have simply left them as global  
  values and referenced them as needed, but instead I passed them     
  explicitly to the non-io routines that needed them.  I could just   
  as easily have declared them here and made them external to the     
  io package, but since file names and IO are so intimately tied, I   
  thought it more logical to declare these variables with the other   
  IO code.                                                            
----------------------------------------------------------------------
*/
extern char  net_config[];                 /* these three from netio.c  */
extern char  net_iop[];
extern char  net_fwt[];
extern char  net_pwt[];
extern char  no_interrupt;
extern char  IO_str[MAX_LINE_SIZE];

static Net  *the_net;          /* variable which holds ptr to the net   */
                               /* (global to this netmain routine only) */

/*
======================================================================
  ROUTINES IN NETMAIN.C
======================================================================
  The routines in this file are grouped below by function.  NO ROUTINE
  ARE PREFIXED IN THIS FILE. 

  The main routine is just below.  It basically amounts to a read-eval-
  print loop, much like an interpreter. After initialization, this code
  loops forever, calling "print_menu", "read_choice", and then 
  "evaluate".  The only other routine is a "check_net_ptr" routine 
  used during evaluate to verify that a valid net exists before 
  attempting some operation.
======================================================================
*/


main()
{
   void  initialize(), print_menu(), evaluate(), cleanup();
   char  read_choice();
   char  c;
      
#if LSPEED
	sys_SetUpVBLTask();
#endif

   signal (SIGINT, T_signal_handle);

   initialize();
   print_menu();
   while (TRUE) {
      c = read_choice();
      if (c == 'q') {
         cleanup();
         break;
      }
      evaluate(c);
   }
#if LSPEED
	sys_RemoveVBLTask();
#endif

}   /* main */


void initialize()
/*
-----------------------------------------------------------------
 This is the initialization routine for the main program. Much   
  of what the program is able to do is assumed here. For example 
  I have assumed (for now) that we will be dealing with only one 
  net at a time (even though I have made allowances in the NET   
  structure for multiple nets).  Thus, this guy only initializes 
  the one global 'the_net' variable.  This would need changing   
  in the future if more nets were added.                         
 Note also that the system random number generator is setup here 
  This is significant since IT MAY NOT BE PORTABLE.              
-----------------------------------------------------------------
*/
{
   the_net = NULL;
#if  USE_SCALED_INTS
   
	A_initialize();
#endif
   sys_init_rand();
   PA_initialize();
   D_initialize();
   
} /* initialize */


void print_menu()
/*
----------------------------------------------------------------------
 This routine prints out the variety of choices available to the user 
  for operating on/creating his net.                                  
----------------------------------------------------------------------
*/
{
   sprintf(IO_str, "\n\n");
   IO_print(0);
   sprintf(IO_str, "NASA-JSC Software Technology Branch\n");
   IO_print(0);
   sprintf(IO_str, "NETS Back Propagation Simulator Version 3.0\n\n");
   IO_print(0);
   sprintf(IO_str, "b -- show bias values\n");
   IO_print(0);
   sprintf(IO_str, "c -- create a net\n");
   IO_print(0);
   sprintf(IO_str, "d -- set dribble parameters\n");
   IO_print(0);
   sprintf (IO_str, "e -- display error stats for a file\n");
   IO_print(0);
   sprintf(IO_str, "g -- generate delivery code\n");
   IO_print(0);
   sprintf(IO_str, "i -- setup I/O pairs\n");
   IO_print(0);
   sprintf(IO_str, "j -- reset I/O pairs\n");
   IO_print(0);
   sprintf(IO_str, "l -- change learning rate\n");
   IO_print(0);
   sprintf(IO_str, "m -- print this menu\n");
   IO_print(0);
   sprintf(IO_str, "n -- show net configuration\n");
   IO_print(0);
   sprintf(IO_str, "o -- reorder I/O pairs for training\n");
   IO_print(0);
   sprintf(IO_str, "p -- propagate an input through the net\n");
   IO_print(0);
   sprintf(IO_str, "r -- reset weights from a file\n");
   IO_print(0);
   sprintf(IO_str, "s -- save weights to a file\n");
   IO_print(0);
   sprintf(IO_str, "t -- teach the net\n");
   IO_print(0);
   sprintf (IO_str, "u -- scale I/O pairs\n");
   IO_print(0);
   sprintf (IO_str, "v -- scale a data file\n");
   IO_print(0);
   sprintf(IO_str, "w -- show weights between two layers \n");
   IO_print(0);
   sprintf(IO_str, "q -- quit program\n");
   IO_print(0);

} /* print_menu */


char  read_choice()
/*
----------------------------------------------------------------------
 Now, this routine will seem simple, but I wanted to separate it out  
  in the event that the future brings more sophisticated input.  Also 
  this program needs to make sure that the input read in is legal,    
  which is really not part of the print_menu function at all.         
----------------------------------------------------------------------
*/
{
   char  result[MAX_LINE_SIZE];

   sprintf(IO_str, "\nNETS Choice(m = menu)? ");
   IO_print(0);
   gets(result);
   while (result[0] != 'b' 
          && result[0] != 'c' 
          && result[0] != 'd' 
          &&result[0] != 'e'
          && result[0] != 'g' 
          && result[0] != 'i' 
          && result[0] != 'j' 
          && result[0] != 'l' 
          && result[0] != 'm' 
          && result[0] != 'n' 
          && result[0] != 'o' 
          && result[0] != 'p' 
          && result[0] != 'r' 
          && result[0] != 's' 
          && result[0] != 't' 
          && result[0] != 'w' 
          && result[0] != 'u'
          && result[0] != 'v'
          && result[0] != 'q') {
      sprintf(IO_str, "\nSorry, I don't understand that input.\n");
      IO_print(0);
      sprintf(IO_str, "Please try again: ");
      IO_print(0);
      gets(result);
   }
   return(result[0]);

} /* read_choice */


void  evaluate(input)
char  input;
/*
----------------------------------------------------------------------
 Evaluates the input command and calls the appropriate function to    
  carry out the actions                                               
----------------------------------------------------------------------
*/
{
   int    t1, t2, check_net_ptr();
   float  tf1;
   char   filename[MAX_WORD_SIZE], file2[MAX_WORD_SIZE], 
          tmp_str[MAX_WORD_SIZE];
   FILE   *fp,*fp2; /* for query from file */

   switch (input) {
      case 'b': {
         if (check_net_ptr(the_net) == OK) {
            if (the_net->use_biases == FALSE) {
               sprintf(IO_str, "\n*** network has no biases");
               IO_print(0);
            } /* end if */
            else {
               sprintf(IO_str, "\n   layer number? ");
               IO_print(0);
               t1 = IO_my_get_int();
               S_show_biases(the_net, t1);
            } /* end else */
         } /* end if */
         break;
      } /* end case */
      case 'c': {
         the_net = B_free_net(the_net);
         IO_set_filenames();
         the_net = B_create_net(1, net_config);
break;
      } /* end case */
      case 'd': {
         D_dribble_status();
         break;
      } /* end case */
      case'e': {
         if(check_net_ptr(the_net) == OK) 
			sprintf (IO_str, " Enter name of file containing I/O pairs for testing: ");
			IO_print(0);
			gets(filename);
         	T_display_error_stats(filename,the_net);
         break;
         } /* end case */
      case 'g': {
         if (check_net_ptr(the_net) == OK) {
            /*--------------------------------*/
            /* use the net_config filename as */
            /* the configuration file.        */
            /*--------------------------------*/
             strcpy(filename, net_config);
            /*-----------------------------*/
            /* prompt for the weights file */
            /*-----------------------------*/
            sprintf(IO_str, "\n   Enter filename with PORTABLE weight values");
            IO_print(0);
            sprintf(IO_str, " (default=%s): ", net_pwt);
            IO_print(0);
            IO_my_get_string(file2);
            if (file2[0] == ENDSTRING) 
               strcpy(file2, net_pwt);

            /*-------------------------*/
            /* call the create routine */
            /*-------------------------*/
            CC_create_delivery(the_net, filename, file2);
         } /* end if */
         break;
      } /* end case */
      case 'i': {
         	if (check_net_ptr(the_net) == OK) {
            	IO_get_io_name();
            	PA_setup_iopairs(the_net, net_iop);
      			if (the_net->scale_file_ptr != NULL)
   					SC_scale_IOpairs (the_net);
   				else
   				{
					sprintf (IO_str,"Do you wish to scale these I/O pairs? ");
					IO_print(0);
					IO_my_get_string(IO_str);
					if ((IO_str[0] == 'y') || (IO_str[0] == 'Y'))
   						SC_scale_IOpairs (the_net);
   				}
			} /* end if */
         break;
      } /* end case */
      case 'j': {
         if (check_net_ptr(the_net) == OK) {
            sprintf(IO_str, "\n   Resetting I/O pairs to 'workfile.net'");
            IO_print(0);
            PA_reset_iopairs(the_net, "workfile.net");
         } /* end if */
         break;
      } /* end case */
      case 'l': {
         if (check_net_ptr(the_net) == OK) {
            sprintf(IO_str, "\n   Layer number: ");
            IO_print(0);
            t1 = IO_my_get_int();
            L_modify_learning(t1, N_get_layer(the_net, t1));
         } /* end if */
         break;
      } /* end case */
      case 'm' : {
         print_menu();
         break;
      } /* end case */
      case 'n': {
         if(check_net_ptr(the_net) == OK)
            S_show_net(the_net);
         break;
      } /* end case */
      case 'o': {
         if(check_net_ptr(the_net) == OK)
            if (the_net->num_io_pairs <= 0) {
               sprintf(IO_str, "\n\n*** no valid set of io pairs to randomize ***\n");
               IO_print(0);
            } /* end if */
            else {
               sprintf(IO_str, "\n   Enter filename for reordered IO pairs: ");
               IO_print(0);
               IO_my_get_string(filename);
               while (filename[0] == ENDSTRING) {
                  sprintf(IO_str, "\n      try again: ");
                  IO_print(0);
                  IO_my_get_string(filename);
               } /* end while */
               PA_randomize_file(filename, "workfile.net", the_net->num_io_pairs,
                                 the_net->num_inputs + the_net->num_outputs);
            } /* end else */

         break;
      } /* end case */
      case 'p': {
         if (check_net_ptr(the_net) == OK) {
            /*----------------------------------*/
            /* Get the input file/screen source */
            /* and number of IO pairs to prop.  */
            /*----------------------------------*/
            sprintf(IO_str, "\n   Enter filename with INPUT(default=manual entry): ");
            IO_print(0);
            IO_my_get_string(filename);
            
            if (filename[0] != ENDSTRING) {
               sprintf(IO_str, "\n   Enter number of I/O pairs to propagate (default=all): ");
               IO_print(0);
               t1 = IO_get_default_int(-1);
            } /* end if */
            
            /*----------------------------------------*/
            /* Get the output file/screen destination */
            /*----------------------------------------*/
            sprintf(IO_str, "\n   Enter OUTPUT destination file(default=screen): ");
            IO_print(0);
            IO_my_get_string(file2);
            if (file2[0] == ENDSTRING)
               fp = NULL;
            else 
               fp = fopen(file2, WRITE_TEXT);
            
            /*-------------------------------*/
            /* call the query function, then */
            /* tidy up output file if used   */
            /*-------------------------------*/
            N_query_net(the_net, filename, fp, t1);
            if (fp != NULL) fclose(fp);
         } /* end if */
         break;
      } /* end case */
      case 'r': {
         if (check_net_ptr(the_net) == OK) {
            sprintf(IO_str,"\n   Load weights from FAST or PORTABLE format");
            IO_print(0);
            sprintf(IO_str, "(f/p, default=f)? ");
            IO_print(0);
            IO_my_get_string(tmp_str);
            if ((tmp_str[0] == 'p') || (tmp_str[0] == 'P'))
               t1 = PORTABLE_FORMAT;
            else t1 = FAST_FORMAT;
            sprintf(IO_str, "\n   Enter name of file with weight values");
            IO_print(0);
            IO_get_wts_name(t1);
            t1 = ( (t1 == FAST_FORMAT) 
                   ? N_reset_wts(the_net, net_fwt, t1)
                   : N_reset_wts(the_net, net_pwt, t1) );
            if (t1 == ERROR) {
               sprintf(IO_str, "\n*** weight resetting was incomplete ***");
               IO_print(0);
            } /* end if */
         } /* end if */
         break;
      } /* end case */
      case 's': {
         if (check_net_ptr(the_net) == OK) {
            sprintf(IO_str,"\n   Save weights in FAST or PORTABLE format");
            IO_print(0);
            sprintf(IO_str, "(f/p, default=f)? ");
            IO_print(0);
            IO_my_get_string(tmp_str);
            if ((tmp_str[0] == 'p') || (tmp_str[0] == 'P'))
               t1 = PORTABLE_FORMAT;
            else t1 = FAST_FORMAT;
            sprintf(IO_str, "\n   Enter file name for storing weights");
            IO_print(0);
            IO_get_wts_name(t1);
            if (t1 == FAST_FORMAT)
               N_save_wts(the_net, net_fwt, t1);
            else N_save_wts(the_net, net_pwt, t1);
         } /* end if */
         break;
      } /* end case */
      case 't': {
         if(check_net_ptr(the_net) == OK) {
            if (the_net->num_io_pairs <= 0) {
               sprintf(IO_str, "\n\n*** no valid set of io pairs to teach ***\n");
               IO_print(0);
            } /* end if */
            else {
               sprintf(IO_str, "\n   Enter constraint error: ");
               IO_print(0);
               tf1 = IO_my_get_float();
               sprintf(IO_str, "\n   Enter max number of cycles(default=%d): ", MAX_CYCLES);
               IO_print(0);
               t1  = IO_get_num_cycles();
               sprintf(IO_str, "\n   Enter cycle increment for showing errors");
               IO_print(0);
               sprintf(IO_str, "(default=1): ");
               IO_print(0);
               t2  = IO_get_default_int(1);
               T_teach_net(the_net, C_float_to_Sint(tf1), t1, t2);
            } /* end else */
         } /* end if */
         break;
      } /* end case */
      case ('u'):
      {
         if (check_net_ptr(the_net)==OK)
         {
	         if (the_net->num_io_pairs <= 0)
	         {
	            sprintf (IO_str," \n\n***Must load I/O pairs prior to scaling***\n\n\n");
	            IO_print(0);
	         } /* end if */
	         else
	         SC_scale_IOpairs(the_net);
	     }
         break;
          }/* end case */
      case ('v'):
      {
         if (check_net_ptr(the_net)==OK)
         {
	         if (the_net->num_io_pairs <= 0)
	         {
	            sprintf (IO_str," \n\n***Must have scaled I/O pairs loaded first***\n\n\n");
	            IO_print(0);
	            break;
	         } /* end if */
	         if (the_net->scale_action == 0)
	         {
	            sprintf (IO_str," \n\n***Must scale I/O pairs first***\n\n\n");
	            IO_print(0);
	            break;
	         } /* end if */
			sprintf (IO_str, " Enter name of file to be scaled> ");
			IO_print(0);
			gets(filename);
			if ((fp=fopen(filename,READ_TEXT))==NULL)
			{
				sprintf (IO_str, " *** File %s didn't open for input ***\n", filename);
				IO_print(0);
				break;
			} /* end while */
			sprintf (IO_str, " Enter name of new scaled file> ");
			IO_print(0);
			gets(filename);
			while ((fp2=fopen(filename,READ_TEXT))!=NULL)
			{
				fclose (fp2);
				sprintf (IO_str, " Warning!  File %s will be overwritten\n", filename);
				IO_print(0);
				sprintf (IO_str, " Press carriage return to confirm or enter new file name> ");
				IO_print(0);
				if (!strlen(gets(filename)))
					break;
			} /* end while */   
			while ((fp2=fopen(filename,WRITE_TEXT))==NULL)
			{
				sprintf (IO_str, " *** File %s didn't open for output ***\n", filename);
				IO_print(0);
				sprintf (IO_str, " Enter new file name or press carriage return to exit> ");
				IO_print(0);
				if (!strlen(gets(filename)))
					break;
			} /* end while */
			if (!strlen(filename))
				break;
			SC_scale_any_file(the_net,fp,fp2);
			fclose(fp);
			fclose(fp2);
         } /* end if */
         break;
          }/* end case */
      case 'w': {
         if (check_net_ptr(the_net) == OK) {
            sprintf(IO_str, "\n   source layer? ");
            IO_print(0);
            t1 = IO_my_get_int();
            sprintf(IO_str, "\n   target layer? ");
            IO_print(0);
            t2 = IO_my_get_int();
            S_show_weights(the_net, t1, t2);
         } /* end if */
         break;
      } /* end case */
      default :
         break;
   } /* end switch */

} /* evaluate */


void  cleanup()
/*
----------------------------------------------------------------------
   Before quitting the program entirely, this routine ensures that all
   the loose ends are accounted for. Right now, that consists of freeing
   all the memory currently used for the network.
----------------------------------------------------------------------
*/
{

   the_net = B_free_net(the_net);
   
} /* cleanup */

#endif

#if MAC_GRAPH

/* ====================================

  MacNETS: A Macintosh-specific version of NETS 3.0

  All input to be propagated must be in a file--no manual entry allowed.
  
   ==================================== */ 

 
/*
----------------------------------------------------------------------
  INCLUDE FILES
----------------------------------------------------------------------
*/

#include "Teach.h"
#include "Net_Options.h"
#include "Get_Layer_.h"
#include "Change_Dribble.h"
#include "About.h"
#include "GraphWind.h"
#include "MenuNames.h"

/* ====================================== */
/* External routines that are called */

extern void InitMyMenus(void);
extern void HandleMenu(short theMenu, short theItem);
extern void A_initialize();
extern void sys_init_rand();
extern void T_signal_handle ();
extern void PA_initialize();
extern void D_initialize();
extern Net    *B_free_net();
extern int doGraph;
extern int yesPrint;
extern int showIOVals;

int defaultNet;			/* flag which indicates if defaults are 	*/
						/* automatically used when creating a net 	*/
int ScreenWidth;
int ScreenHeight;
Net  *the_net;          /* variable which holds ptr to the net   */
int t1, t2; 
int homeVolume;			/* the home directory of the nets application */
int doUI;				/* Flag which indicates whether or not the UI is active during propagation */
char doneFlag;			/* Has the user decided to quit? */
char	propFileName[MAX_WORD_SIZE];
float maxerrList[NUMERRS], rmserrList[NUMERRS];		/* the error history */

/* ====================================== */
void main(void);
void welcome(void);
void HandleEvent(EventRecord *myEvent);
void DoMacUI(void);
/* ====================================== */


void main()
{
	int 		 hsize, vsize, wleft, wtop, i;
	EventRecord  myEvent;
	Str255  sTemp;
	
    void  initialize(), cleanup();
	
    signal (SIGINT, T_signal_handle);
    
	InitGraf(&thePort);						/* Initialize Mac toolbox thingys */
	InitFonts();
  	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs(NIL);

	MaxApplZone();
	FlushEvents(everyEvent,0);
  	
    InitMyMenus();							/* Set up the menus */
	Init_GraphWind();						/* Initialize Net Error window */
	Init_NetInputWind();					/* Initialize Net Input window */
	Init_NetOutputWind();					/* Initialize Net Output window */
	
	doneFlag = FALSE;
	defaultNet = TRUE;
	doUI = TRUE;
	t1 = 0;
	t2 = 2;
	
	doGraph = 1;         					/* Do error graph when teaching */
	yesPrint = 0;         					/* Don't print errors when teaching */
	showIOVals = 0;         				/* Don't do input and output window updating when teaching */

	/* Set the checkmarks on the menus for the above flags */
	CheckItem(GetMHandle (Net_Functions_Menu),Item_DoGraph,doGraph);          
	CheckItem(GetMHandle (Net_Functions_Menu),Item_YesPrint,yesPrint);
	CheckItem(GetMHandle (Net_Functions_Menu),Item_ShowIOVals,showIOVals);
	
	propFileName[0] = '\0';
	GetVol(&sTemp,&homeVolume);						/* Get the home volume */
	
	WaitNextEvent(everyEvent,&myEvent,0L,NULL);		/* Brings application to foreground under some systems */
	
    ScreenHeight = screenBits.bounds.bottom - screenBits.bounds.top;  /* Get screen sizes */
    ScreenWidth = screenBits.bounds.right - screenBits.bounds.left;

	InitCursor();							/* Set cursor to arrow */			

	D_About();								/* Display the About dialog */

	for( i=0; i<(NUMERRS); i++)
		maxerrList[i] = rmserrList[i] = 1.0;		/* Initialize the error arrays */
	
    Open_GraphWind(0.2);					/* Create the Net Error window with a constraint line at 0.2 */
    
	wleft = SCREENLEFT;						/* Set up the dimensions of the Net Info window */
    wtop = SCREENTOP + 160+22;
    hsize = 474;
    vsize = (ScreenHeight - (wtop + 6));
    if (vsize < 126) 
    	vsize = 126;
    CreateDisplayWindow(hsize,vsize,wleft,wtop);	/* Create the Net Info window */
	
	SetCursor(*GetCursor(watchCursor));		/* Set cursor to Watch */
    initialize();							/* Do Nets initialization */
    welcome();								/* Print welcome to window */
    
   	do 										/* Main loop */
   		DoMacUI();							/* Handle user interface */
	while (doneFlag == FALSE);				/* Are we done? */
	
    cleanup();								
	return;

}

/*
 *	DoMacUI
 *
 *	Checks to see if there is an event, and calls the event handler if there is one.
 *	No parameters, no return value.
 *
 */
 
void DoMacUI()
{
	EventRecord  myEvent;
 
 	if (WaitNextEvent(everyEvent,&myEvent,0L,NULL))	
		HandleEvent(&myEvent);
}


/*
 *	HandleEvent
 *
 *	A big "switch" statement to handle the different events
 *	A pointer to the event record is passed in as a parameter.
 *
 */

void HandleEvent(EventRecord *myEvent)
{		
	char         Is_A_Dialog;
	char         stillInGoAway;
	char         ch;
	short        code;
	short        theMenu,theItem;
	short        chCode;
	long         mResult;
	WindowPtr    whichWindow;
	Rect         tempRect,OldRect;
	Point        myPt;
	GrafPtr      SavePort;

	code = FindWindow((*myEvent).where, &whichWindow);	
	switch ((*myEvent).what)
	{
		case mouseDown:
			switch (code)
			{
			  /*====================================*/
			  /* Handle click in desk - do nothing. */
			  /*====================================*/
			  
			  case inDesk: 
				break;
			
			  /*===========================*/
			  /* Handle click in menu bar. */
			  /*===========================*/
			  
			  case inMenuBar:								
				mResult = MenuSelect((*myEvent).where);
				theMenu = HiWord(mResult);
				theItem = LoWord(mResult);
   				SetCursor(*GetCursor(watchCursor));
				HandleMenu(theMenu,theItem);				/* Dispatch the menu handler */
				break;
 
			  /*==============================*/
			  /* Handle click in drag region. */
			  /*==============================*/

			  case inDrag:
			    DoDrag(whichWindow,myEvent);
				break;
 
			  /*==============================*/
			  /* Handle click in size region. */
			  /*==============================*/
			  
			  case inGrow:
				DoGrow(whichWindow,myEvent);
				break;
 
 
			  /*=================================*/
			  /* Handle click in content region. */
			  /*=================================*/
			  
			  case inContent:
				DoContent(whichWindow,myEvent);
				break;
 
			  /*================================*/
			  /* Handle click in system window. */
			  /*================================*/

			  case inSysWindow:
				SystemClick(myEvent, whichWindow);
				break;

		 	  /*=================================*/
			  /* Handle click in zoom in region. */
			  /*=================================*/

		      case inZoomIn:
		        DoZoom(whichWindow,inZoomIn,myEvent);
		        break;
		        
			  /*==================================*/
			  /* Handle click in zoom out region. */
			  /*==================================*/
			  
		      case inZoomOut:
		        DoZoom(whichWindow,inZoomOut,myEvent);
		        break;

			}
			break;
			
		case keyDown: 	/* Key press handler and */
		case autoKey: 	/* Key repeat handler */
			ch = (*myEvent).message &  charCodeMask;
			if ((*myEvent).modifiers & cmdKey)			/* if command-key is held down during */
			{											/* keypress, then handle as a menu choice */
				mResult = MenuKey(ch);
				theMenu = HiWord(mResult);
				theItem = LoWord(mResult);
				if (theMenu != 0) 
					HandleMenu(theMenu, theItem); 		/* Dispatch menu handler */
			}
			break;

		case updateEvt:
			whichWindow = (WindowPtr)(*myEvent).message;
			switch (GetWRefCon(whichWindow)) 
			{ 
				case 0:
 					UpDate_GraphWind();
					break;

				case 1:
 					UpDate_NetInputWind();
					break;

				case 2:
 					UpDate_NetOutputWind();
					break;

				default:
					DoUpdate(myEvent);
					break;
			}
			break;

		case diskEvt:
			if (HiWord((*myEvent).message) != 0) 
			{
				(*myEvent).where.h = ((screenBits.bounds.right - screenBits.bounds.left) / 2) - (304 / 2);
				(*myEvent).where.v = ((screenBits.bounds.bottom - screenBits.bounds.top) / 3) - (104 / 2);
				InitCursor();
				theItem = DIBadMount((*myEvent).where, (*myEvent).message);
			}
			break;


		case activateEvt:
			DoActivate(myEvent);
			break;
			
 	    /*==============================*/
	    /* Handle suspend/resume event. */
	    /*==============================*/
	    case app4Evt:
	    	DoSuspendResume(myEvent);
			break; 


		default:
			break;
 
	}
 
	InitCursor();
}


/*
-----------------------------------------------------------------
 initialize()
 This is the initialization routine for the main program. Much   
  of what the program is able to do is assumed here. For example 
  I have assumed (for now) that we will be dealing with only one 
  net at a time (even though I have made allowances in the NET   
  structure for multiple nets).  Thus, this guy only initializes 
  the one global 'the_net' variable.  This would need changing   
  in the future if more nets were added.                         
 Note also that the system random number generator is setup here 
  This is significant since IT MAY NOT BE PORTABLE.              
-----------------------------------------------------------------
*/

void initialize()
{
   the_net = NULL;
#if  USE_SCALED_INTS
   
	A_initialize();
#endif
   sys_init_rand();
   PA_initialize();
   D_initialize();
   
} /* initialize */


/*
----------------------------------------------------------------------
 cleanup()
   Before quitting the program entirely, this routine ensures that all
   the loose ends are accounted for. That consists of freeing
   all the memory currently used for the network, and closing windows
   on the Macintosh.
----------------------------------------------------------------------
*/

void  cleanup()
{

	the_net = B_free_net(the_net);
	Close_GraphWind();
	Close_NetInputWind();
 	Close_NetInputWind();
	Close_NetOutputWind();
   
} /* cleanup */


/*
 * welcome() 
 *
 * Print the introductory message
 *
 */ 

void welcome()
{
	extern char  IO_str[MAX_LINE_SIZE];
    sprintf(IO_str, "NASA-JSC Software Technology Branch\n");
    IO_print(0);
    sprintf(IO_str, "NETS Back Propagation Simulator, Version 3.0m (Macintosh)\n");
    IO_print(0);
	sprintf(IO_str, "\nUnlike regular NETS, this window is used only for output. MacNETS uses a\n");
	IO_print(0);
	sprintf(IO_str, "regular Macintosh interface, i.e., commands are entered via the Menu Bar.\n");
	IO_print(0);
	sprintf(IO_str, "\nREMINDER:\n");
	IO_print(0);
	sprintf(IO_str, "Default values are automatically used when creating a net. To change this, \n");
	IO_print(0);
	sprintf(IO_str, "uncheck the \"Use Defaults with Create Net\" item under the \"Net Setup\" menu.\n");
	IO_print(0);

}
#endif


int  check_net_ptr(ptr_net)
Net  *ptr_net;
/*
----------------------------------------------------------------------
 This small routine is simply a check to see whether or not the arg   
  passed in is actually a valid Net pointer.  If a net has not been   
  created, or if there was an error in its creation, then no          
  operations ought to be allowed using the net pointer.               
 Returns 0 if the ptr is INvalid, otherwise returns a 1.  Note that   
  a net which has an error will return a non-NULL pointer, but the ID 
  of such a net will be negative. (see return values of B_create_net) 
----------------------------------------------------------------------
*/
{
   if (ptr_net == NULL) {
      sprintf(IO_str, "\n\n*** no valid net exists ***\n");
      IO_print(0);
      return(0);
   }
   if (ptr_net->ID == ERROR) {
      sprintf(IO_str, "\n\n*** no valid net exists ***\n");
      IO_print(0);
      return(0);
   }
   return(OK);

} /* check_net_ptr */

