/* Copyright 1997 Lucent Technologies Inc. - All Rights Reserved */

#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/param.h>
#include <fcntl.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <ctype.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <limits.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "arg.h"
#include "stage.h"
#include "client.h"
#include "crypto.h"
#include "util.h"

char	root[MAXPATHLEN];	/* Where the root of his file system is, locally */

int debug = 0;	/* debug */
int nflg = 0;	/* don't make any changes */
int qflg = 0;	/* quick: skip cryptographic checksum */
int sflg = 0;	/* show stats */
int vflg = 0;	/* verbose - show protocol */
int only = 0;	/* only update directories, not their subdirectories */

int cmd_count = 0;	/* commands processed */
int files_checked = 0;
int files_transferred = 0;
int bytes_transferred = 0;
int responses = 0;
int response_wait = 0;

char *
get_root(void) {
	char *ep;
	char *cp = strtok(NULL, "	 \r\n");

	if (cp == NULL)
		return "/";

	/*
	 * Remove a trailing /, if any.
	 */
	ep = strrchr(root, '/');
	if (ep && (ep-root+1) == strlen(root))
		*ep = '\0';
	return cp;
}

/*
 * Delete a remote file.
 */
int
delete(int fd, char *path) {
	remcmd(fd, "rm", remotepath(path));
	remresp(fd);
	if (!remok)
		return 2;
	return 1;
}

int
sendfile(int rfd, char *lpath, char *rpath) {
	int i, n, fd;
	char buf[MAXIO];
	char *er;

	remcmd(rfd, "pu", rpath);
	remcmd(rfd, "++", statinfo(lpath));
	remresp(rfd);
	if (!remok) {
		fprintf(stderr, "could not create file %s: %s\n",
			rpath, remerr);
		return 0;
	}
	fd = open(lpath, O_RDONLY);
	if (fd < 0) {
		fprintf(stderr, "could not open %s\n", lpath);
		return 0;
	}
	if (vflg > 1)
		fprintf(stderr, "<--- (%d bytes)\n", laststat.st_size);

	er = writenetfile(fd, rfd, laststat.st_size);
	if (er) {
		fprintf(stderr, "Error transferring %s: %s, aborting\n",
			lpath, er);
		exit(13);
	}
	close(fd);
	remresp(rfd);
	if (!remok) {
		fprintf(stderr, "could not put file %s: %s\n",
			rpath, remerr);
		return 0;
	}
	files_transferred++;
	bytes_transferred += laststat.st_size;
	return 1;
}

/*
 * Update the remote file, if needed.  Return 1 if we actually had
 * to do it, 2 if update failed.
 */
int
check_file_update(int fd, char *path, int new) {
	char *mess;
	struct stat st, *stp;
	int rc;

	files_checked++;
	if (new)
		stp = 0;
	else
		stp = get_remote_stat(fd, path);
	if (stp == 0) {
		/* remote file not present */
		mess = "cr %s";
		goto sendfile;
	}

	if (S_ISDIR(stp->st_mode)) {	/* remote file is a directory */
		if (!nflg && delete(fd, path) > 1) {
			fprintf(stderr, "file update: could not delete remote file/dir\n");
			return 2;
		}
		mess = "rdf %s";
		goto sendfile;
	}

	/*
	 * Both local and remote are a file.  Compare, first the easy stuff,
	 * then with a checksum to see if we need to update it.
	 */
	rc = stat(path, &st);
	if (rc != 0) {
		fprintf(stderr, "local file %s disappeared\n", path);
		return 2;
	}
	if (st.st_mtime == stp->st_mtime && st.st_size == stp->st_size) {
		char *lcs;
		mode_t remmode =
			((stp->st_mode & 0777) & ~srv_umask) | srv_or_perm;
		mode_t locmode =
			  ((st.st_mode & 0777) & ~srv_umask) | srv_or_perm;
		if (locmode != remmode) {
			/*
			 * Permissions have changed.  Be lazy, and recopy
			 * the file.
			 */
			mess = "cm %s";
			goto sendfile;
		}
			
		/*
		 * We don't care if the permission bits are different,
		 * which is wrong.  In fact, it is wrong, but we
		 * need to know the remote umask to do it right.
		 */
		if (qflg)
			return 0;
		/* 
		 * Files might be the same. Compute and compare checksums.
		 */
		remcmd(fd, "cs", remotepath(path));
		lcs = checksum(path);
		remresp(fd);
		if (!remok) {
			/*
			 * remote file disappeared.  Create it.
			 */
			mess = "cr %s";
			goto sendfile;
		}
		if (strlen(lcs) == sizeof(DIGEST)*2 &&
		    strlen(response) == sizeof(DIGEST)*2 &&
		    strcmp(lcs, response) == 0)
			return 0;
	}
	mess = "cp %s";

sendfile:
	if (!nflg) {
		if (!sendfile(fd, path, remotepath(path)))
			return 2;
	}	
	if (vflg || nflg)
		log(mess, remotepath(path));
	return 1;
}

/*
 * Update the remote directory recursively.  Return 1 if any changes were needed
 * to the directory or any directory below it. Return 2 if any errors occurred.
 */
int
check_dir_update(int fd, char *path, int new) {
	struct stat *st;
	char *lp, *lls, *rls, *ep, *cp;
	char *mess;
	int changed = 0, err = 0, newdir = 0;

	lls = dirls(path);
	if (new)
		st = 0;
	else
		st = get_remote_stat(fd, path);

	if (st == 0) {
		newdir = 1;	/* remote file/dir not present */
		mess = "md %s";
	} else if (S_ISREG(st->st_mode)) {
		/*
		 * If the remote item is a file, remove it, and create a fresh
		 * directory.
		 */
		if (!nflg && delete(fd, path) > 1) {
			free(lls);
			return 2;
		}
		mess = "rfd %s";
		newdir = 1;
	} else {
		/*
		 * The remote directory exists.  Make a string listing the local
		 * directory entries, then get a list of the remote directory
		 * names, and delete those that are not present in the local
		 * directory.
		 */
		rls = get_remote_ls(fd, path);
		if (rls == 0) {
			fprintf(stderr, "check_dir_update, rls failure\n");
			free(lls);
			return 2;
		}
		for (cp=rls; cp[1] != '/'; cp=ep) {
			char nc;

			ep = strchr(cp+1, '/');
			if (ep == 0) {
				fprintf(stderr, "check_dir_update: inconceivable\n");
				exit(13);
			}
			nc = ep[1];	/* save next char */
			ep[1] = '\0';
			if (strstr(lls, cp) == NULL) { /* not found locally */
				char fn[MAXPATHLEN];
				ep[0] = '\0';
				strcpy(fn, remotepath(path));
				strcat(fn, cp);
				if (!nflg) {
					remcmd(fd, "rm", fn);
					remresp(fd);
					if (!remok) {
						fprintf(stderr, "check_dir_update: remote file not deleted, %s\n", remerr);
						err = 1;
					}
				}
				changed = 1;
				if (vflg || nflg)
					fprintf(stderr, "rm %s\n", fn);
				ep[0] = '/';
			}
			ep[1] = nc;
		}
		free(rls);
	}

	if (newdir) {
		char *rp = remotepath(path);

		if (*rp != 0) {
			if (!nflg) {
				remcmd(fd, "md", rp);
				remcmd(fd, "++", statinfo(path));
				remresp(fd);
				if (!remok) {
					fprintf(stderr, "could not create remote directory %s: %s\n",
						remotepath(path), remerr);
					free(lls);
					return 2;
				}
			}
			changed = 1;
			if (vflg || nflg)
				log(mess, rp);
			new = 1;
		}
	}

	/*
	 * Now, update each entry that is supposed to be there.
	 */
	for (lp=lls; lp[1] && lp[1] != '/'; *ep = '/', lp = ep) {
		char fn[MAXPATHLEN];
		struct stat st;

		ep = strchr(lp+1, '/');
		*ep = '\0';
		strcpy(fn, path);
		strcat(fn, lp);
		if (debug > 1)
			fprintf(stderr, "*** checking %s\n", fn);
		if (stat(fn, &st) < 0) {
			fprintf(stderr, "Local file %s inaccessible (%s)\n",
				path, strerror(errno));
			err = 1;
			continue;
		}
		if (S_ISDIR(st.st_mode)) {
			if (!only)
				changed |= check_dir_update(fd, fn, new);
		} else
			changed |= check_file_update(fd, fn, new);
	}
	free(lls);
	return (changed|err);
}

int
usage(void) {
	fprintf(stderr,
		"usage: stage [-c config] [-h host] [-k keyfile] [-p port] [-q] [-s] [-v] [-n] service [path ...]\n");
	return(1);
}

int
main(int argc, char *argv[]) {
	time_t	t0, t1;
	int srvfd;
	char *hostname = 0;
	char *service = 0;
	int serverport = SERVERPORT;
	char wd[MAXPATHLEN];
	int rc = 0;
	char *csrvfile = CSRVFILE;
	char *keyfile = 0;
	char *ep, *host;
	KEY key;

	ARGBEGIN {
	case 'c': csrvfile = ARGF();		break;
	case 'd': debug++;			break;
	case 'h': hostname = ARGF();		break;
	case 'k': keyfile = ARGF();		break;
	case 'n': nflg++;			break;
	case 'o': only++;			break;
	case 'p': serverport = atoi(ARGF());	break;
	case 'q': qflg++;			break;
	case 's': sflg++;			break;
	case 'v': vflg++;			break;
	default:
		exit(usage());
	} ARGEND;

	if (argc < 1)
		exit(usage());

	service = *argv++;
	argc--;

	host = get_service_host(csrvfile, service);
	if (hostname == 0)
		hostname = host;

	strcpy(root, get_root());
	if(debug)
		fprintf(stderr, "root=%s\n", root);

	get_key(keyfile, service, key);

	if ((srvfd = tcpconnect(hostname,serverport)) < 0) {
		fprintf(stderr, "stage: connecting to %s:%d, service %s: %s\n",
			hostname, serverport, service, strerror(errno));
		exit(2);
	}

	ep = start_client_crypto(srvfd, key, service);
	if (ep) {
		fprintf(stderr, "stage: %s, %s: %s\n", hostname, service, ep);
		exit(3);
	}
	if (debug)
		log("crypto setup complete");

	if (sflg)
		t0 = rtime();

	/*
	 * Drop any special setuid/setgid privs they may have given us to
	 * read the key file.
	 */
	if (setuid(getuid()) < 0) {
		perror("stage: setuid");
		exit(13);
	}
	if (setgid(getgid()) < 0) {
		perror("stage: setgid");
		exit(13);
	}
	if (getcwd(wd, MAXPATHLEN) == 0) {
		fprintf(stderr, "stage: getwd: %s\n", wd);
		exit(13);
	}
	if (argc == 0) {
		argv[0] = root;
		argc = 1;
	}

	get_server_info(srvfd);

	for (; argc; argv++, argc--) {
		struct stat st;
		char path[MAXPATHLEN*2 + 1];

		if (strlen(argv[0]) > MAXPATHLEN-1) {
			fprintf(stderr, "path too long: %s\n", argv[0]);
			continue;
		}
		if (argv[0][0] == '/')
			strncpy(path, argv[0], sizeof(path)-1);
		else {
			/*
			 * Convert relative path to absolute.  Directory
			 * '.' is ugly in the logs: don't include it.
			 */
			strcpy(path, wd);
			if (strcmp(argv[0], ".") != 0) {
				strcat(path, "/");
				strcat(path, argv[0]);
			}
		}
		if (stat(path, &st) < 0) {
			if (errno == ENOENT) {
				/*
				 * We could just issue a delete, but checking
				 * makes the logging right.
				 */
				if (get_remote_stat(srvfd, path)) {
					if (!nflg)
						rc |= delete(srvfd, path);
					if (vflg || nflg)
						log("rm %s", remotepath(path));
				}
				continue;
			} else {
				perror("stat local file");
				exit(1);
			}
		}
		if (S_ISDIR(st.st_mode))
			rc |= check_dir_update(srvfd, path, 0);
		else
			rc |= check_file_update(srvfd, path, 0);
	}
	remcmd(srvfd, "ex", "");
	remresp(srvfd);

	if (sflg) {
		double t;
		t1 = rtime();

		t = t1 - t0;
		fprintf(stderr, "%.2f cmds/s,  %.2f files/s,  %d updates,  %.1f ms ave resp.\n",
			cmd_count/(t/1000.0), files_checked/(t/1000.0),
			files_transferred, 
			(double)response_wait/(double)responses);
	}
	return rc;
}
