/* Write an MPI/C program that computes a correlation function between two sets 
  of values. This should have a similar form to the Nbody problem, but without 
  the complication of using something like Runge-Kutta to compute gravity between
  each pair of bodies. Instead there will be a much simpler function to compute 
  for each pair of "bodies", f(a,b) = abs ( a*a + b*b ).
  Much Better Message Parallel Algorithm and the ring topology are used.
  @auther Hyun Kim
  @date 03/29/00
*/

#include "mpi.h"

double f( double x, double y )
{
	return x*x + y*y;
}

void main( int argc, char **argv )
{
	double local_value = 0,	/* each processor sends its value to the process 0 */
		  total = 0,		/* total value */
		  x[1000], y[1000], correlation[1000];
	int source, dest, stag = 0, rtag = 0,
		my_rank, num_processes, local_range,
		i, j, k;
	double begin_time, stop_time;
	MPI_Status status;

	/* start up MPI
	*/
	MPI_Init( &argc, &argv );

	/* set up the starting time
	*/
	begin_time = MPI_Wtime();
  
	/* get process rank
	*/
	MPI_Comm_rank( MPI_COMM_WORLD, &my_rank );

	/* get number of processes
	*/
	MPI_Comm_size( MPI_COMM_WORLD, &num_processes );

	/* find the source/destination of each process for a ring topology
	  and the range for each process
	*/
	dest = ( my_rank + 1 ) % num_processes;
	source = ( my_rank - 1 + num_processes ) % num_processes;
	local_range = 1000 / num_processes;

	/* Each process sends its value to the next processor
	*/
	for( i = my_rank*local_range; i < (my_rank+1)*local_range; i++ )
	{
		correlation[i] = 0;		/* initialize the correlation */
		x[i] = i * 0.003; 
		y[i] = i * 0.007;
	}
	for( k = 0; k < num_processes; k++ )
	{
		/* compute the correlation of x's and y's in a processor */
		for( j = my_rank*local_range; j < (my_rank+1)*local_range; j++ )
		{
			for( i = my_rank*local_range; i < (my_rank+1)*local_range; i++ )
			{
				correlation[i] += f( x[i], y[j] );
			}
		}
		/* communicate y's */
		MPI_Sendrecv_replace( &y[my_rank*local_range], local_range, MPI_DOUBLE, dest, stag, source, rtag, MPI_COMM_WORLD, &status );
	}

	/* sum the correlation in each processor */
	for( i = my_rank*local_range; i < (my_rank+1)*local_range; i++ )
	{
		local_value += correlation[i];
	}

	/* each processor sends its result to processor 0 */
	if( my_rank != 0 )
	{
		MPI_Send( &local_value, 1, MPI_DOUBLE, 0, stag, MPI_COMM_WORLD );
	}
	else
	{
		total = local_value;
		for( i = 1; i < num_processes; i++ )
		{
			MPI_Recv( &local_value, 1, MPI_DOUBLE, i, rtag, MPI_COMM_WORLD, &status );
			total += local_value;
		}
	}

	/* stop the time
	*/
	stop_time = MPI_Wtime();

	/* print out the result
	*/
	if( my_rank == 0 )
	{
		printf( "Elapsed_Time\t Total_Value\n" );
		printf( "%f\t %f\n", stop_time - begin_time, total );
	}

	/* close MPI
	*/
	MPI_Finalize();
}
