#include "mpi.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define PI 3.141592654
#define FROM 0
#define TO PI*0.5
#define TRAPEZOIDS 128 

/* function prototypes
*/

/* calculate the local integral
*/
double compute_integral( double from, double to, int how_many, double width );

/* compute the function which is approximated for integration
*/
double my_function( double x );

/* main
*/
int main( int argc, char** argv )
{
	int my_rank, num_processes, how_many, i;
	double trapezoid_width, local_integral, total_integral, local_from, local_to;
	int tag = 0, dest = 0;
	MPI_Status status;

	/* start up MPI
	*/
	MPI_Init( &argc, &argv );

	/* get process rank
	*/
	MPI_Comm_rank( MPI_COMM_WORLD, &my_rank );

	/* get number of processes
	*/
	MPI_Comm_size( MPI_COMM_WORLD, &num_processes );

	/* width of the local interval (trapezoid)
	*/
	trapezoid_width = ( TO - FROM ) / TRAPEZOIDS;

	/* number of trapezoids per process
	*/
	how_many = TRAPEZOIDS / num_processes;

	local_from = FROM + my_rank * how_many * trapezoid_width;
	local_to = local_from + how_many * trapezoid_width;
	local_integral = compute_integral( local_from, local_to, how_many, trapezoid_width );

	/* Each process of which rank is not 0 prints its contribution, 
	  then sends its local integral to the destination process 0.
	   Process 0 sums all local integrals from other processes.
	*/
	printf( " Process %d: contribution = %f\n", my_rank, local_integral );
	if( my_rank == 0 )
	{
		total_integral = local_integral;
		for( i = 1; i < num_processes; i++ )
		{
			MPI_Recv( &local_integral, 1, MPI_FLOAT, i, tag, MPI_COMM_WORLD, &status );
			total_integral += local_integral;
		}
	}
	else
	{
		MPI_Send( &local_integral, 1, MPI_FLOAT, dest, tag, MPI_COMM_WORLD );
	}

	/* print the total integral
	*/
	if( my_rank == 0 )
	{
		printf( "\n Total Integration: %f\n", total_integral );
	}

	/* close MPI
	*/
	MPI_Finalize();

	return 0;
}

/* function implementation
*/
double compute_integral( double from, double to, int how_many, double width )
{
	double my_integral = 0.0, left, right;
	int i;

	left = my_function( from );
	for( i = 1; i <= how_many; i++ )
	{
		right = my_function( from + width * ( double )i );
		my_integral += width * 0.5 * ( left + right );
		left = right;
	}

	return my_integral;
}
   
double my_function( double x )
{
	double y = cos( x );
	return ( 1.0 / sqrt( 1.0 - 0.5 * y * y ) );
}