#include <mpi.h>
#include <stdlib.h>
#include <math.h>

#define MAX(X,Y) ((X) > (Y) ? (X) : (Y))
#define MIN(X,Y) ((X) > (Y) ? (Y) : (X))
#define SQR(X)   ((X)*(X))
#define ABS(X)   ((X>=0) ? X : -(X))

/* this define a fish data structure */
typedef struct
{
  double px, py;  /* position */
  double vx, vy;  /* velocity */
  double mass;    /* mass */
} fish_t;

int main (int argc, char *argv[])
{
  /* MPI datatypes */
  MPI_Datatype MPI_fish;
  MPI_Status   status;
  int          my_size, my_rank;

  /* fishes related data types */
  fish_t *fishes;
  int     num_fishes;
  int    *my_first_fish;
  int    *my_last_fish;
  double *force_x;
  double *force_y;

  double dt_min, dt_max, dx_min, grav_const, end_time;

  /* temp variables */
  int    i,j;
  double dt;
  double time;
  int int_time;
  double my_max_sq_velocity, max_sq_velocity;

  /* initialize the MPI environment */
  MPI_Init(&argc, &argv);

  /* create MPI_fish type */
  MPI_Type_contiguous(5, MPI_DOUBLE, &MPI_fish);
  MPI_Type_commit(&MPI_fish);

  /* get number of processors */
  MPI_Comm_size(MPI_COMM_WORLD, &my_size);

  /* get my processor id */
  MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);

  /* initialize and report simulation parameters */
  num_fishes = 50;
  dt_min = 0.0001;
  dt_max = 0.01;
  dx_min = 0.005;
  grav_const = 1.0;
  end_time = 10.0;

  if (my_rank == 0)
  {
    printf("procs                  = %d\n", my_size);
    printf("fish                   = %d\n", num_fishes);
    printf("min dt                 = %f\n", dt_min);
    printf("max dt                 = %f\n", dt_max);
    printf("min dx                 = %f\n", dx_min);
    printf("gravitational constant = %f\n", grav_const);
    printf("simulation end         = %f\n", end_time);
    fflush(stdout);
  }

  /* allocate memory for all the fishes */
  fishes = (fish_t *)malloc(num_fishes * sizeof(fish_t));

  /* allocate memory for holding the start and end position
     of fishes for each processor */
  my_first_fish = (int *)malloc(my_size * sizeof(int));
  my_last_fish  = (int *)malloc(my_size * sizeof(int));

  /* and initialize them */
  my_first_fish[0] = 0;
  my_last_fish[0]  = num_fishes / my_size;
  for (i = 1; i < my_size; i++)
  {
    my_first_fish[i] = my_last_fish[i - 1] + 1;
    my_last_fish[i]  = my_last_fish[i - 1] + num_fishes / my_size;
  }
  my_last_fish[my_size - 1] = num_fishes - 1;

  /* allocate memory for the forces acting on each fish */
  force_x = (double *)malloc(num_fishes * sizeof(double));
  force_y = (double *)malloc(num_fishes * sizeof(double));

  /* initialize my fishes location, velocity and mass */
  for (i = my_first_fish[my_rank]; i <= my_last_fish[my_rank]; i++)
  {
      double z       = 2.0 * M_PI / num_fishes;
      fishes[i].px   = sin(z * i);
      fishes[i].py   = cos(z * i);
      fishes[i].vx   = 0;
      fishes[i].vy   = 0;
      fishes[i].mass = 1.0 / num_fishes;
  }

  /* loop over each time step */
  dt = dt_min;
  for (time = 0; time < end_time; time += dt)
  {

	printf("\n\nTime = %d\n",time);
    /* let everyone know about all the fishes in the world */
    if (my_rank == 0)
    {
      /* gather all fishes information from everyone */
      for (i = 1; i < my_size; i++)
      {
        MPI_Recv(&fishes[my_first_fish[i]],
                 my_last_fish[i] - my_first_fish[i] + 1,
                 MPI_fish, i, 100, MPI_COMM_WORLD, &status);
      }
      /* then, boardcast to everyone */
      MPI_Bcast(fishes, num_fishes, MPI_fish, 0, MPI_COMM_WORLD);
    }
    else
    {
      /* send my fishes information to rank 0 */
      MPI_Send(&fishes[my_first_fish[my_rank]],
               my_last_fish[my_rank] - my_first_fish[my_rank] + 1,
               MPI_fish, 0, 100, MPI_COMM_WORLD);
      /* then, receive all fishes information from rank 0 */
      MPI_Bcast(fishes, num_fishes, MPI_fish, 0, MPI_COMM_WORLD);
    }

    /* compute the forces act on my fishes */

    for (i = my_first_fish[my_rank]; i <= my_last_fish[my_rank]; i++)
    {

      /* initialize the forces to zero */
      force_x[i] = 0;
      force_y[i] = 0;

      /* for each fishes in the world */
      for (j = 0; j < num_fishes; j++)
      {

        /* calculate the forces act on fishes[i] based on the distance */
        double dx      = fishes[j].px - fishes[i].px;
        double dy      = fishes[j].py - fishes[i].py;

        /* enforces the minimun distance */
        double sq_dist = MAX (SQR(dx) + SQR(dy), SQR(dx_min));

        /* calculate and accumulate the forces from fishes[j] */
        double grav_base =
          grav_const * fishes[i].mass * fishes[j].mass / sq_dist;
        force_x[i] += dx * grav_base;
        force_y[i] += dy * grav_base;
      }
    }

    /* compute the new positions of my fishes */

    /* calculate my fishes velocity */
    my_max_sq_velocity = 0;
    for (i = my_first_fish[my_rank]; i <= my_last_fish[my_rank]; i++)
    {
      fishes[i].px += fishes[i].vx * dt;
      fishes[i].py += fishes[i].vy * dt;
      fishes[i].vx += force_x[i] / fishes[i].mass * dt;
      fishes[i].vy += force_y[i] / fishes[i].mass * dt;
      my_max_sq_velocity = MAX(my_max_sq_velocity,
                               SQR(fishes[i].vx) + SQR(fishes[i].vy));
    }

    /* get the maximum fish velocity from everyone */
    MPI_Allreduce (&my_max_sq_velocity,
                   &max_sq_velocity,
                   1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);

    /* recompute dt to bound the movement of next time step */
    dt = dx_min / sqrt(max_sq_velocity);
    dt = MIN(dt_max, dt);
    dt = MAX(dt_min, dt);

    int_time = time/2;
    /* output fish positions every ten timesteps */
    if(time/2.0 == (double)int_time)
    {
		for (i = 0; i < num_fishes; i++)
		{
			printf("fish%d(%f, %f)\n",i,(fishes[i].px + 1) / 2 * 0.5 + 0.25,(fishes[i].py + 1) / 2 * 0.5 + 0.25);
		}
    }
  }

  /* release all memory allocated before exit MPI */
  free(fishes);
  free(my_first_fish);
  free(my_last_fish);
  free(force_x);
  free(force_y);

  /* now, exit MPI */
  MPI_Barrier(MPI_COMM_WORLD);

  MPI_Finalize();
}