/*

n-Queens Problem

Command line arguments:

  - number of queens
  - threshold
  - [m] (optional, masks printing board)

Threshold is the number of rows up to which master works.
Number of queens effectively sets the size of the board.
(board size = no. queens * no. queens)

Master should generate one subproblem at a time, not recursively
solve the entire subproblems at once. dfs() returns at a leaf node,
saving the current internal nodes at the stack, which directs the next
search.

Master places queens up to the given threshold row, and worker starts
from the threshold row.

Load balancing is roughly measured by summing up the number of nodes
visited for each processors, i.e. by summing up the number of calls to
place_queen().

*/

#include <stdio.h>
#include <mpi.h>
#include <string.h>
#include <math.h>

#define ANY MPI_ANY_SOURCE
#define ALL MPI_COMM_WORLD
#define TAG 1

#define NOQUEEN 8
#define QUEEN 1

#define ABOVE 0
#define LEFT -1
#define RIGHT +1

#define TRUE 1
#define FALSE 0
#define DONE -2

#define on_board(row, col) (row>=0 && row<brd->nqueen && col>=0 && col<brd->nqueen)
#define last_row(row) (row == brd->nqueen - 1)
#define queen_in(row, col) (brd->board[row * brd->nqueen + col] == QUEEN)

#define bd(i, j) (brd->board[(i) * brd->nqueen + (j)])
#define on_thres(row, col) (row>=0 && row<brd->thres && col>=0 && col<brd->nqueen)
#define thres_row(row) (row == brd->thres - 1)

typedef struct _Board {

  int id;              /* processor id */
  int *board;          /* chess board */
  int nqueen;          /* number of queens - board size */
  int thres;           /* threshold */
  int *stack;
  int *rq_list;        /* request queue */
  int **wk_list;       /* work queue */
  int rq_head;
  int rq_tail;
  int rq_size;
  int wk_head;
  int wk_tail;
  int wk_size;
  int num_node;        /* visited nodes count */
  int mask;            /* masks printing board */
} Board;

Board* board_create(int id, int nprocs, int nqueen, int thres, int mask) {

  Board *brd;
  int i, size;

  size = nqueen * nqueen;
  brd = (Board*)malloc(sizeof(Board));
  brd->id = id;
  brd->board = (int*)malloc(sizeof(int) * size);
  brd->nqueen = nqueen;
  brd->thres = thres;
  for (i=0; i<size; i++)
    brd->board[i] = NOQUEEN;
  brd->num_node = 0;
  brd->mask = mask;
  brd->rq_list = NULL;
  brd->wk_list = NULL;
  if (!nprocs) return brd;

  /* below are info only for master */
  brd->stack = (int*)malloc(sizeof(int) * thres);
  size = (int)pow(nqueen, thres);
  brd->rq_size = 2 * size;
  if (2 * size < 2 * nprocs) brd->rq_size = 2 * nprocs;
  brd->rq_list = (int*)malloc(sizeof(int) * brd->rq_size);
  brd->wk_size = size;
  brd->wk_list = (int**)malloc(sizeof(int*) * brd->wk_size);
  brd->rq_head = brd->rq_tail = brd->wk_head = brd->wk_tail = -1;

  for (i=0; i<thres; i++)
    brd->stack[i] = 0;

  return brd;
}

void req_add(Board *brd, int num) {

  brd->rq_tail++;
  brd->rq_list[brd->rq_tail] = num;
}

void work_add(Board *brd) {

  int *work, size;

  size = brd->nqueen * brd->nqueen;
  work = (int*)malloc(sizeof(int) * size);
  memcpy(work, brd->board, sizeof(int) * size);  /* copy board */
  brd->wk_tail++;
  brd->wk_list[brd->wk_tail] = work;
}

int req_remove(Board *brd) {

  int retval;

  brd->rq_head++;
  retval = brd->rq_list[brd->rq_head];
  return retval;
}

int* work_remove(Board *brd) {

  int *retval;

  brd->wk_head++;
  retval = brd->wk_list[brd->wk_head];
  return retval;
}

int req_avail(Board *brd) {

  if (brd->rq_head < brd->rq_tail) return 1;
  return 0;
}

int work_avail(Board *brd) {

  if (brd->wk_head < brd->wk_tail) return 1;
  return 0;
}

void board_destroy(Board *brd) {

  free(brd->board);
  free(brd->stack);
  if (brd->rq_list != NULL)
    free(brd->rq_list);
  if (brd->wk_list != NULL)
    free(brd->wk_list);
  free(brd);
}

void show_board(Board *brd) {

  int row, col, element;
  if (brd->mask) return;  /* board printing masked */
  printf("\nsolution, process %d\n", brd->id);
  for (row=0; row<brd->nqueen; row++) {
    for (col=0; col<brd->nqueen; col++)
      if (bd(row, col) == QUEEN)
        printf("Q");
      else
        printf("-");
    printf("\n");
  }
}

int no_queen(Board *brd, int slant, int row, int column) {

  int result;

  column = column + slant;
  row--;
  if (on_board(row, column))
    if (queen_in(row, column)) result = FALSE;
    else result = no_queen(brd, slant, row, column);
  else
    result = TRUE;
  return result;
}

void dfs(Board *brd, int row, int column) {

  brd->num_node++;
  /* searches for a leaf node of the tree */
  if (on_thres(row, column)) {
    if (no_queen(brd, ABOVE, row, column) &&
        no_queen(brd, LEFT, row, column) &&
        no_queen(brd, RIGHT, row, column)) {
      bd(row, column) = QUEEN;
      brd->stack[row] = column;   /* save current position for next search */
      if (thres_row(row)) {
        work_add(brd);
        bd(row, column) = NOQUEEN;
        return;
      }
      else dfs(brd, row+1, 0);    /* search until threshold */
    }
    else dfs(brd, row, column+1); /* try other column */
  }
  else {
    brd->stack[row] = DONE;       /* this row is done, go up the tree */
    if (row > 0 && brd->stack[row-1] < brd->nqueen)
      bd(row-1, brd->stack[row-1]) = NOQUEEN;
  }
}

void place_queen(Board *brd, int row, int column) {

  brd->num_node++;
  /* searches for the remaining tree */
  if (on_board(row, column)) {
    if (no_queen(brd, ABOVE, row, column) &&
        no_queen(brd, LEFT, row, column) &&
        no_queen(brd, RIGHT, row, column)) {
      bd(row, column) = QUEEN;
      if (last_row(row)) show_board(brd);
      place_queen(brd, row+1, 0);
      bd(row, column) = NOQUEEN;
    }
    place_queen(brd, row, column+1);
  }
}

void manager(int id, int nprocs, int nqueen, int thres, int mask) {

  int count, *nextbrd, i, done, msg, size;
  Board *brd;
  MPI_Status status;

  brd = board_create(id, nprocs, nqueen, thres, mask);
  count = nprocs - 1;
  size = nqueen * nqueen;
  done=FALSE;
  dfs(brd, 0, 0);
  for (;;) {
    /* generate subproblem */
    if (done == FALSE) {
      /* find where the previous search was */
      for (i=thres-1; i>0 && brd->stack[i]==DONE; i--) ;
      brd->stack[i]++;
      dfs(brd, i, brd->stack[i]);
      if (brd->stack[0] == DONE) done = TRUE;  /* root node is done */
    }
    /* add request to the queue */
    MPI_Iprobe(ANY, TAG, ALL, &msg, &status);
    if (msg == TRUE) {
      MPI_Recv(&msg, 1, MPI_INT, ANY, TAG, ALL, &status);
      req_add(brd, status.MPI_SOURCE);
    }
    /* remove a request from queue and send work */
    if (req_avail(brd) && work_avail(brd)) {
      nextbrd = work_remove(brd);
      MPI_Send(nextbrd, size, MPI_INT, req_remove(brd), TAG, ALL);
      free(nextbrd);
    }
    /*signal end of work */
    if (req_avail(brd) && done==TRUE && !work_avail(brd)) {
      brd->board[0] = DONE;
      MPI_Send(brd->board, size, MPI_INT, req_remove(brd), TAG, ALL);
      count--;
    }
    if (count == 0) break;
  }

  printf("process%3d, number of node visited%8d\n", id, brd->num_node);
  board_destroy(brd);
}

void worker(int id, int nqueen, int thres, int mask) {

  Board *brd;
  int size;
  MPI_Status status;

  size = nqueen * nqueen;
  brd = board_create(id, 0, nqueen, thres, mask);

  for (;;) {
    MPI_Send(&id, 1, MPI_INT, 0, TAG, ALL);   /* send request */
    MPI_Recv(brd->board, size, MPI_INT, 0, TAG, ALL, &status);
    if (brd->board[0] == DONE) break;         /* empty msg, no more work */
    place_queen(brd, thres, 0);               /* work */
  }
  printf("process%3d, number of node visited%8d\n", id, brd->num_node);
  board_destroy(brd);
}

void main(int argc, char **argv) {

  int id, size, nqueen, thres, mask;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(ALL, &size);
  MPI_Comm_rank(ALL, &id);

  if (argc > 2) {
    nqueen = atoi(argv[1]);
    thres = atoi(argv[2]);
    if (argc > 3 && *argv[3] == 'm')
      mask = 1;
    else mask = 0;
  }
  else {
    mask = 0;
    nqueen = NOQUEEN;
    thres = 3;
  }

  if (id == 0) manager(id, size, nqueen, thres, mask);
  else worker(id, nqueen, thres, mask);

  MPI_Finalize();
}
