/*--------------------------------------------------------------------------------
 * ntype.c
 * FILE: ntype.c
 * DESCRIPTION:
 *             This program computes a correlation function 
 *             between two sets of measurements x & y of size
 *             N.  The measurements are equally distributed between
 *             the processors.  The "Much Better Message Parallel 
 *             Algorithm is used in which we compute:
 *
 *             for each x[i] correlation[i] = Sum{0<=j<N}(f(x[i],y[j]))
 *
 *             Note the usage of the MPI_Sendrecv function which is
 *             used to form the ciruclar chain between the processors
 *             in such a way that the y's from each processor in the 
 *             communicator visits each other processor exactly once.
 *             Also, note the usage of MPI_Reduce to efficiently compute
 *             the final correlation function between the two sets of 
 *             measurements x & y.
 * Author: Sameet Kudyadi, PHY-403, 3-5981
 * Version: 1.0
 * Date: 11/06/98
 -------------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>

#define N 8

int funct(int m1, int m2);

void main(int argc, char* argv[])

{ 
        int n;    /*used for evenly distributing the N values to each processor  */
        int sendto;    /*rank of process that the current process is sending to  */
        int recvfrom;    /*  rank of process from which current process receives data  */     
        int my_rank;  /*  rank of the current process  */     
        int x[N], y[N];  /*  arrays containing the two sets of measurements  */       
        int local_correlation[N];    /*  correlation array for each processor initialized to zero  */
	int temp[N];    /*  temporary array used to receive y's from other processes  */
        int sum;  /*  used to sum up the elements of the correlation array for each processor  */
        int correlation;    /*  the sum of all the correlation functions calculated by each process  */
        int p;  /*  the total number of processes in the communicator  */
        int i,j,k;    /*  used for looping  */

	MPI_Comm grid_comm;    /*  communicator that will contain all processes in MPI_COMM_WORLD  */  
	MPI_Status status;
                
        /*  Initializes MPI environment using command line arguments  */
        MPI_Init(&argc, &argv);

        /*  Finds out how many processes are being used  */
        MPI_Comm_size(MPI_COMM_WORLD, &p);

        /*  Gets the rank of the process  */
	MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);

        n = N/p;       /*  each processor has N/p values  */      
        
        /* assigning random variables to x & y and initializing arrays */
        for (i=0; i<n;i++) 
	{
	  y[i] = my_rank*n + i;
	  x[i] = my_rank*n + i;
	  local_correlation[i] = 0;
	  temp[i]=0;
	}
        sum = 0; correlation=0;   /*  initializing to zero  */
	for (k=0; k<p; k++) 
	{	
 
	  for ( j=0; j<n; j++)
	    for ( i=0; i<n; i++)             
	      local_correlation[i] += funct(x[i], y[j]);
                               
	  sendto = (my_rank+1);
	  if (sendto >= p)
	    sendto = 0;
	  recvfrom = my_rank-1;
	  if (recvfrom < 0)
	    recvfrom = p-1;
		
	  MPI_Sendrecv(&y, n, MPI_INT, sendto ,0, &temp, n, MPI_DOUBLE, recvfrom ,0, MPI_COMM_WORLD, &status); 
                  
	  for(i = 0; i < n; i++)	
	    y[i] = temp[i];              
	}

     
        /*   summing up the elements of the correlation array  by each process  */
        for(i = 0; i < n; i++)
	  sum += local_correlation[i];
			

        /*  a global reduction operation carried out to sum up the correlation functions obtained from all the processors  */
	MPI_Reduce(&sum, &correlation, 1, MPI_INT, MPI_SUM, 0, MPI_COMM_WORLD);

        /*  Process 0 displays the correlation function obtained  */
	if (my_rank == 0)
	  printf("\nThe correlation function is %d\n\n", correlation); 

	MPI_Finalize();
}

/*  The function that is used to compute the correlation  */
int funct(int m1, int m2)
{
  return (m1*m1 + m2*m2);
}
