/*---------------------------------------------------------------------------------------------------
 * pisquared.c
 * FILE: pisquared.c
 * DESCRIPTION:
 *             This program calculates the value of PI squared 
 *             using the "trapezoidal rule" in two dimensions.  
 *             It uses the cartesian coordinate system to split
 *             the interval 0 to 1 of the function, into square grids.
 *             It assigns each processor to a grid.  Functions in the
 *             mpi.h header are used to create this Cartesian Grid and
 *             also to get the coordinates of each grid that is assigned
 *             to each processor.  Each process then computes the integral 
 *             of that grid.  In this way, all the processes contribute to the 
 *             calculation of pi squared.  This version uses MPI_Gather to 
 *             gather up individual integrals of each process.  The master 
 *             process (process 0) sums up the integrals to give the value of PI^2.  
 * Author: Sameet Kudyadi, PHY-403, 3-5981
 * Version: 1.0
 * Date: 10/29/98
 ------------------------------------------------------------------------------------------------------*/ 


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>

double sqrt();

main(int argc, char* argv[])
{
       double PI =3.14159265358979323842643;
       int p;        /*  The number of processes  */
       int my_grid_rank;       /*  The rank of each processor assigned to a Cartesian Grid  */
       double h;        /*  Trapezoidal base length  */
       int q;        /*  square root of the number of processes - used to construct the square grid  */
       MPI_Comm grid_comm;   /*  communicator that will contain all processes in MPI_COMM_WORLD  */  
       int dim_sizes[2];  /*  array that specifies the order of each dimension - the number of rows and colums */
       int wrap_around[2];  /*  array that specifies whether each dimesion is circular  */
       int reorder = 1;  /*  reorders the processes in the new communicator */
       int coordinates[2];  /*  array that gives the coordinates of any process in the grid  */
       double startx;  /*  starting point of interval , along x direction */
       double starty;   /*  starting point of interval, along y direction */
       double endx;  /*  endpoint of interval, along x direction  */
       double endy;  /*  endpoint of interval, along y direction  */
       double integral;  /*  Integral calculated by each process */
       double pisquared;  /*  the result of the computation of pi^2  */
       int i,j; /*  used for looping  */
       double total[10];  /*  buffer containing each sub-integral  */
       double Trapezoid(double local_start_x, double local_start_y, double local_end_x, double local_end_y, int local_n, double h);
                    /* function used to calculate the local integral  */

       /*  Initializes MPI environment using command line arguments  */
       MPI_Init(&argc, &argv);

       /*  Finds out how many processes are being used  */
       MPI_Comm_size(MPI_COMM_WORLD, &p);

       q = sqrt(p);  /*  gets the square root of the number of processes to be used for the rows and colums  */
       dim_sizes[0] = dim_sizes[1] = q;  /*  sets the order of each dimension to the number of rows and colums respectively  */
       wrap_around[0] = wrap_around[1] = 1; /*  allows wraparound i.e each dimension is circular  */
          
       /*  Creates a new communicator grid_comm by caching a Cartesian topology with MPI_COMM_WORLD  */
       MPI_Cart_create(MPI_COMM_WORLD, 2, dim_sizes, wrap_around, reorder, &grid_comm);
       
       /*  Gets the rank of the process so that it's coordinates can be obtained by a call to the next function MPI_Cart_coords  */ 
       MPI_Comm_rank(grid_comm, &my_grid_rank);
        
       /*  Gets the coordinates of a process in the communicator using the rank of the process  */
       MPI_Cart_coords(grid_comm, my_grid_rank, 2, coordinates);
        
       h = 1/(double)p;  /*  h is the same for all processes  */
         
       startx = (double)coordinates[0]/(double)q;  /*  start point coordinate for each process grid in x-direction  */
       starty = (double)coordinates[1]/(double)q;  /*  start point coordinate for each process grid in y-direction  */
       endx = (double)coordinates[0]/(double)q + (1/(double)q); /*  end point coordinate for each process grid in x-direction  */
       endy = (double)coordinates[1]/(double)q + (1/(double)q); /*  end point coordinate for each process grid in x-direction  */

       integral = Trapezoid(startx, starty, endx, endy, q, h);  /*  local integral is calculated  */

       /*  Each process sends it's integral to master process 0  */
       /*  Process 0 stores the integrals in rank order in array total  */
       MPI_Gather(&integral, 1, MPI_DOUBLE, &total, 1, MPI_DOUBLE, 0, grid_comm);

        /*  Master process performs the summation to find PI  */
        /*  It broadcasts pi to all other processes  */
        /*  It displays the value of pi  */
        if(my_grid_rank == 0)
        { 
               pisquared = 0.0; 
               for(i = 0; i <= p; i++)
                        pisquared += total[i];
               printf("The value of pi squared is %.16f\n", pisquared); 
               printf("The value of pi is %.16f\n", sqrt(pisquared)); 
               printf("The error between computer value and actual value of pi is %.16f\n", fabs(sqrt(pisquared) - PI));
        }         
        MPI_Finalize();
}


/*---------<  Trapezoid is used to calculate the local integral  >--------*/
double Trapezoid(double local_start_x, double local_start_y, double local_end_x, double local_end_y, int local_n, double h)
{
	double integ;	/*  used to store the integral calculated by each process  */
	double x,y;
	int i,j;
	
	double f(double x, double y);	/*  function that we are integrating over  */
                
                /*  end point functions are calculated and summed  */
	integ = (f(local_start_x, local_end_y) + f(local_end_x, local_start_y) + f(local_start_x, local_start_y) + f(local_end_x,local_end_y))/4.0;
	
                x = local_start_x;
                y = local_start_y;
                if (local_n!=1)  /*  if number of processes is 1, then no need to go into these computations  */
                    for (i = 0; i<=local_n; i++)       /*  x-direction loop  */
  	    {              
	                for (j = 0; j <= local_n; j++)     /*  y-direction loop  */
		{
                                        if (((x!=local_start_x) || (x!=local_end_x)) ||  ((y!=local_start_y) || (y!=local_end_y)))  /*  ignores end points of the grid  */
			integ = integ + f(x,y)/2.0;
                                        y=y+h;    
		}
                                x=x+h;
	    }          
	integ = integ*h*h;
	return integ;
}

/*---------<   Calculates the given function  >-----------*/
double f(double x, double y)
{
	double return_val;
	/*  Calculating the function 16/(1 + x*x)(1 + y*y)  */
	return_val = (16)/(( 1 + x*x )*(1 + y*y));
	return return_val;
}	
	


